{
  Roland Potter - Woodstock Computer Solutions
  http://wcsweb.co.uk

  This code is provided for educatonal purposes only
  with no warranty whatsoever.

  Feel free to adapt it and use it in applcations commercial
  or otherwise and open source componenents but not as part
  of commercially distrbuted developer components.

  Flat File Dataset Component.
  Requires definition file and data file to open,
  or events to load same from a stream.
}
unit wcsffdataset;

interface
uses
  Windows, SysUtils, Classes, WcsDataObjects, WcsFlatFileType,
  DB, wdcDSet, FMTBcd;

const
  cWCSFFReadBuffSize = $10000;
type
  TWCSStreamGetEvent = procedure(Sender: TObject; var Stream: TStream; const ForOutput: Boolean) of object;

  TWCSFlatFileRecord = ^TWCSFlatFileRecordStore;
  TWCSFlatFileRecordStore = record
    fOffset : Int64;     // offset in original stream.
    fRecNo  : Longint;   // record number  in original stream.
    fFltNo  : Longint;   // record number  in original stream.
    fLength : Longint;   // Length in original stream.
  end;

  PFFRecBuff = ^TFFRecBuff;
  TFFRecBuff = array[0..$FFFFFF] of Byte;
  TFFStreamBuff = array[0..cWCSFFReadBuffSize] of Char;

  PWCSFFTransBuff = ^TWCSFFTransBuff;
  TWCSFFTransBuff = record
    TransUsage: TWCSFlatFileUsgType;
    TransNull : Boolean;
    TransSigned: Boolean;
    TransSize : integer;
    TransLen  : integer;
    case TWCSFlatFileUsgType of
      wffu_Chars     : (TransPC   : PChar    );
      wffu_LongInt   : (TransLong : PLongint );
      wffu_LargeInt  : (TransInt64: PInt64   );
      wffu_Float     : (TransFloat: PExtended);
      wffu_BCD       : (TransBCD  : PBCD     );
      wffu_DateTime  : (TransDate : PDateTime);
  end;

  TFFFieldPosArray = array of integer;
  TFFBufferInfo = ^TFFBufferInfoStore;
  TFFBufferInfoStore = record
    fFFBuffRec    : TWCSFlatFileRecord;
    fFFRecBuff    : PFFRecBuff;
    fFFFldBuff    : TWCSFFTransBuff;
    fFFFieldCount : integer; // VOLATILE if delimited!
    fFFFieldOffs  : TFFFieldPosArray;
    fFFFieldLens  : TFFFieldPosArray;
  end;

  TWCSPreFilterEvent = procedure(const RecBufInfo: TFFBufferInfo; var Accept: Boolean) of object;

  TWCSFlatFileDataset = class(TWcsMemListDataset)
  private
    fFFLayoutFile : String;
    fFFDataFile   : String;
    fFFStream     : TStream;
    fFFFreeStream : Boolean;
    //!!!! if too slow with a list and objects - swap over to a > 4 byte List
    fFFRecList    : TList;
    fFFHeadList   : TList;
//    fFFData       : TMemoryStream;
//    fFFRecordData : Pointer;
    fFFRecLen     : integer; // VOLATILE if delimited!
    fFFRecOffs    : integer;
    fFFRecDataLen : integer;
    fFFFieldsMax  : integer;

    fFFRecBuffInfo : TFFBufferInfoStore;
    fFFCmpBuffInfo : TFFBufferInfoStore;

(*
    fFFRecBuff    : PFFRecBuff;
    fFFFldBuff    : TWCSFFTransBuff;
    fFFFieldCount : integer; // VOLATILE if delimited!
    fFFFieldOffs  : TFFFieldPosArray;
    fFFFieldLens  : TFFFieldPosArray;
    // need cmp versions if delimited
    fFFCmpBuff    : PFFRecBuff;
    fFFFldCmpBuff : TWCSFFTransBuff;
    fFFCmpFieldCount : integer; // VOLATILE if delimited!
    fFFCmpFieldOffs  : TFFFieldPosArray;
    fFFCmpFieldLens  : TFFFieldPosArray;
*)

    fFFDateMasks  : array of TDateMaskRec;
    fFFRecType    : TWCSFlatRecType;
    fFFRecDelim   : String;
    fFFRecDelimLen : integer;
    fFFRecDelimCR : Boolean;
    fFFFldDelim   : String;
    fFFFldDelimLen : integer;
    fFFDelimQuotes : TWCSFlatQuoteStore;

    fFFReadBuff: TFFStreamBuff;
    fFFReadBuffPosn : Longint;
    fFFReadBuffLeft : Longint;
    fFFWriteBuff: TFFStreamBuff;
    fFFWriteBuffPosn : Longint;
    fFFWriteBuffLeft : Longint;

    fDefinitions: TFieldDefinitionList;
    fFFIndexFields: String;
    fIndexFieldNos: array of integer;
    fIndexFieldCount: integer;
    fIndexError: Boolean;
    fFFOnGetLayoutStream: TWCSStreamGetEvent;
    fFFOnGetDataStream: TWCSStreamGetEvent;
    fFFOnPreFilter: TWCSPreFilterEvent;
    fTotalRecCount: integer;
    function RecordNoFromPtr(const RecordPtr: Pointer): Integer;
    function IndexCompare(Object1, Object2: TObject): Integer;
    procedure SetFFIndexFields(const Value: String);
    procedure CheckIndexFields(const IFString: String);
    function ReadDelimitedLine(var RecLen, TotLen: Longint): Boolean;
    function FlushIn: Boolean;
    function FlushOut: Boolean;
    function ReadStreamBuff(var Buff; const BuffLen: Longint): Longint;
    function WriteStreamBuff(const Buff; const BuffLen: Longint): Longint;
    procedure AllocLayoutBuffers(const ForFieldCount: integer);
    procedure DeAllocLayoutBuffers;
    function ReadDelimitedFields(const ARecBuffInfo: TFFBufferInfo;
      const OQ, CQ: Char): integer;
    function ReadDelimitedField(const ARecBuffInfo: TFFBufferInfo;
      const FldNo: integer;  const OQ, CQ: Char): integer;
    procedure FldBuffBestType(const Len: integer;
      var StartType: TWCSFlatFileBestScanType);
  protected
    procedure InternalClose; override;
    procedure InternalOpen; override;

    function ListGetField(ListObject : TObject; ListIndex: integer;
                          Field: TField; FieldBuf: Pointer): Boolean; override;
    procedure ListSetupFieldDefs; override;
    procedure ListSetupFields; override;
    procedure ListDataRead; override;
    procedure ListDataFetch; override;
    // This is read only so these all raise exception
    procedure ListFreeElement(const Element: TObject); override;
    function ListCreateElement(const Element: TObject): TObject; override;
    procedure ListSetObject(NewObj: TObject; var NewIdx: Longint); override;
    procedure ListInsert(const NewObj : TObject; var NewIdx: integer); override;
    procedure ListDelete(const DelIdx : integer); override;
    procedure ListSetField(ListObject : TObject;
                           Field: TField; FieldBuf: Pointer); override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    // moved to public and can be called independently (no field usage)
    procedure FreeData;
    procedure ReadData;
    procedure ReadLayout;
    // For delimited files only - returns record count
    function PreScanToLayout(const FldDefs: TFieldDefinitionList): integer;

    procedure DelphivarFromField(const RecBuffInfo: TFFBufferInfo; const FldNo: integer);
    property TotalRecCount: integer read fTotalRecCount;
  published
    property FFDataFile: String read fFFDataFile write fFFDataFile;
    property FFLayoutFile: String read fFFLayoutFile write fFFLayoutFile;
    property FFIndexFields: String read fFFIndexFields write SetFFIndexFields;
    property FFOnGetDataStream: TWCSStreamGetEvent read fFFOnGetDataStream write fFFOnGetDataStream;
    property FFOnGetLayoutStream: TWCSStreamGetEvent read fFFOnGetLayoutStream write fFFOnGetLayoutStream;
    property FFOnPreFilter: TWCSPreFilterEvent read fFFOnPreFilter write fFFOnPreFilter;
  end;

procedure FieldDefinitionToDelphi(const FDef: TFieldDefinition;
  var FldType: TFieldType; var FldSize, FldDec: integer; var DTMask: TDateMaskRec);
procedure ReAllocWCSFFTransBuff(var Buff: TWCSFFTransBuff; const Size: integer);
function WCSFFTransBuffToStr(const Buff: TWCSFFTransBuff): String;
function FieldDefExistsProc(const S: String; const UserInfo: Pointer): Boolean;

implementation
uses
  dbConsts;

type
  TStrExistProc = function(const S: String; const UserInfo: Pointer): Boolean;

const
  DatePartChars = ['0'..'9','A'..'Z','a'..'z'];

function LastCharPos(const FindCh : Char; const SearchStr : String) : Integer;
begin
  for Result := Length(SearchStr) downto 1 do
    if SearchStr[Result] = FindCh then
      Exit;
  Result := 0;
end;

function IncrementalName(const StartName: String;
  const ExistsFunc: TStrExistProc; const UserInfo: Pointer;
  const Numpad: integer = 0; const MaxLen: integer = 255): String;
var
  ni, E, P : integer;
  nBase, FormStr, AddStr: STring;
  AlreadyIncremented: Boolean;
begin
  P := LastCharPos('_', StartName);
  AlreadyIncremented := (P > 0);
  if AlreadyIncremented then
  begin
    Val(Copy(StartName, P+1, Length(StartName) - P), ni, E);
    AlreadyIncremented := E = 0;
  end;
  if AlreadyIncremented then
  begin
    nBase := Copy(StartName,1,P-1);
    inc(ni);
  end else begin
    ni := 2;
    nBase := StartName;
  end;
  if NumPad > 1 then
    FormStr := '_%' + IntToStr(NumPad) + '.' + IntToStr(NumPad) + 'd'
  else
    FormStr := '_%d';
  repeat
    AddStr := Format(FormStr, [ni]);
    E := Length(AddStr);
    if Length(nBase) + E > MaxLen then
      nBase := Copy(nBase,1, MaxLen - E);
    Result := nBase  + AddStr;
    inc(ni);
  until not ExistsFunc(Result, UserInfo);
end;

function FieldDefExistsProc(const S: String; const UserInfo: Pointer): Boolean;
begin
  Result := TDataset(UserInfo).FieldDefs.IndexOf(S) >= 0;
end;

function WCSFFNewRec(const AOffset : Int64; const ARecNo, AFltNo, ALength : Longint): TWCSFlatFileRecord;
begin
  New(Result);
  with Result^ do
  begin
    fOffset := AOffset;
    fRecNo  := ARecNo ;
    fFltNo  := AFltNo ;
    fLength := ALength;
  end;
end;

procedure ReAllocWCSFFTransBuff(var Buff: TWCSFFTransBuff; const Size: integer);
begin
  ReAllocMem(Buff.TransPC, Size);
  Buff.TransSize := Size;
end;

function WCSFFTransBuffToStr(const Buff: TWCSFFTransBuff): String;
begin
  with Buff do
  if TransNull then
    Result := ''
  else
    case TransUsage of
      wffu_LongInt : Result := IntToStr(TransLong^);
      wffu_LargeInt: Result := IntToStr(TransInt64^);
      wffu_Float   : Result := FloatToStr(TransFloat^);
      wffu_DateTime: Result := DateTimeToStr(TransDate^);
      wffu_BCD     : Result := FloatToStr(BCDToDouble(TransBCD^));
      else SetString(Result, TransPC, TransLen);
    end;
end;

procedure FieldDefinitionToDelphi(const FDef: TFieldDefinition;
  var FldType: TFieldType; var FldSize, FldDec: integer; var DTMask: TDateMaskRec);
begin
 // certain FieldTypes have VERY illogical use of size and precision
  with FDef do
  begin
    FldSize := 0;
    FldDec  := 0;
    case DUsgType of
      wffu_Chars    :
        begin
          FldType := ftString;
          FldSize := DSrcSize;
        end;
      wffu_LongInt   :
        begin
          FldType := ftInteger;
        end;
      wffu_LargeInt   :
        begin
          FldType := ftLargeInt;
        end;
      wffu_Float     :
        begin
          FldType := ftFloat;
        end;
      wffu_BCD       :
        begin
        // We can only ever move this data anywhere via a Float (ODBC datasets etc)
        // So for now suppress the BCD...
          FldType := ftFloat;
        {
          FldType := ftBCD;
          FldSize := DSrcDP;
          FldDec  := DSrcPrec;
        }
        end;
      wffu_DateTime  :
        begin
          DtMask := StrToDateMaskRec(DSrcDstMask);
          if DTMask.Posns[dptYear] > 0 then
            if DTMask.Posns[dptHour] > 0 then
              FldType := ftDateTime
            else
              FldType := ftDate
          else
            if DTMask.Posns[dptHour] > 0 then
              FldType := ftTime
            else
              FldType := ftDateTime;
        end;
      wffu_Binary    :
        begin
          FldType := ftBlob;
        end;
    end;
  end;
end;

{ TWCSFlatFileDataset }

constructor TWCSFlatFileDataset.Create(AOwner: TComponent);
begin
  inherited;
  fDefinitions:= TFieldDefinitionList.Create;
  fFFRecList  := TList.Create;
  ExtList     := fFFRecList;
  fFFHeadList := TList.Create;
  ReadOnly := True;
end;

destructor TWCSFlatFileDataset.Destroy;
begin
  fDefinitions.Free;
  inherited;
  fFFRecList.Free;
  fFFHeadList.Free;
end;

function TWCSFlatFileDataset.RecordNoFromPtr(const RecordPtr: Pointer): Integer;
begin
  Result := fFFRecList.IndexOf(RecordPtr);
end;

procedure TWCSFlatFileDataset.CheckIndexFields(const IFString: String);
var
  IndexCheck: String;
  i,j,P, IndexCount : integer;
  OneIndexCheck : String;
  CheckFieldNos: array of integer;
begin
  fIndexError := False;
  if (IFString > '') then
  begin
    if IFString[Length(IFString)] = ';' then
      IndexCheck := Uppercase(IFString)
    else
      IndexCheck := Uppercase(IFString) + ';';
    IndexCount := 0;
    P := pos(';', IndexCheck);
    while P > 0 do
    begin
      inc(IndexCount);
      IndexCheck := Copy(IndexCheck, P+1, 1024);
      P := pos(';', IndexCheck);
    end;
    if IFString[Length(IFString)] = ';' then
      IndexCheck := Uppercase(IFString)
    else
      IndexCheck := Uppercase(IFString) + ';';
    SetLength(CheckFieldNos, IndexCount);
    FillChar(CheckFieldNos[0], IndexCount * Sizeof(Integer), 0);
    IndexCount := 0;
    P := pos(';', IndexCheck);
    while P > 0 do
    begin
      OneIndexCheck := Copy(IndexCheck, 1, P-1);
      for i := 0 to fDefinitions.Count - 2 do
        if CompareText(OneIndexCheck, fDefinitions[i].DName) = 0 then
        begin
          CheckFieldNos[IndexCount] := i+1;
          Break;
        end;
      inc(IndexCount);
      IndexCheck := Copy(IndexCheck, P+1, 1024);
      P := pos(';', IndexCheck);
    end;
    for i := 0 to fIndexFieldCount - 1 do
      if CheckFieldNos[i] <= 0 then
      begin
        fIndexError := True;
        Break;
      end;
    if not fIndexError then
    begin
      fIndexFieldCount := IndexCount;
      SetLength(fIndexFieldNos, IndexCount);
      Move(CheckFieldNos[0], fIndexFieldNos[0], IndexCount * Sizeof(Integer));
    end;
  end else begin
    SetLength(fIndexFieldNos, 0);
    fIndexFieldCount := 0;
  end;
end;

procedure TWCSFlatFileDataset.AllocLayoutBuffers(const ForFieldCount: integer);
begin
  // This is calculated so put into private var
  fFFRecType := fDefinitions.RecordType;
  case fFFRecType of
    wffl_Delimited:
      begin
        fFFRecDelim := fDefinitions.RecDelim;
        fFFRecDelimLen := Length(fFFRecDelim);
        if fFFRecDelimLen = 0 then raise exception.Create('Delimited data must have a record delimiter');
        fFFFldDelim := fDefinitions.FldDelim;
        fFFFldDelimLen := Length(fFFFldDelim);
        if fFFFldDelimLen = 0 then raise exception.Create('Delimited data must have a field delimiter');
        if pos(fFFRecDelim, fFFFldDelim) = 1 then raise exception.Create('Record delimiter cannot be contained in Field delimiter');
        // cater for the either/or unpredictable #13 #10 combinations
        fFFRecDelimCR := (fFFRecDelimLen = 2) and (fFFRecDelim[1] = #13) and (fFFRecDelim[2] = #10);
        fFFDelimQuotes := fDefinitions.DelimQuotes;
      end;
    wffl_Variable : raise exception.Create('Variable length records not currently supported');
    wffl_Parsed   : raise exception.Create('Parsed Text records not currently supported');
  end;

  fFFFieldsMax  := ForFieldCount;
  with fFFRecBuffInfo do
  begin
    fFFFieldCount := ForFieldCount;
    SetLength(fFFFieldOffs, fFFFieldsMax);
    SetLength(fFFFieldLens, fFFFieldsMax);
  end;
  with fFFCmpBuffInfo do
  begin
    fFFFieldCount := ForFieldCount;
    SetLength(fFFFieldOffs, fFFFieldsMax);
    SetLength(fFFFieldLens, fFFFieldsMax);
  end;

  SetLength(fFFDateMasks, fFFFieldsMax);
  FillChar(fFFDateMasks[0], fFFFieldsMax * Sizeof(TDateMaskRec), 0);

end;

procedure TWCSFlatFileDataset.DeAllocLayoutBuffers;
begin
  fFFFieldsMax  := 2;
  with fFFRecBuffInfo do
  begin
    fFFFieldCount := 2;
    SetLength(fFFFieldOffs, fFFFieldsMax);
    SetLength(fFFFieldLens, fFFFieldsMax);
  end;
  with fFFCmpBuffInfo do
  begin
    fFFFieldCount := 2;
    SetLength(fFFFieldOffs, fFFFieldsMax);
    SetLength(fFFFieldLens, fFFFieldsMax);
  end;
  FillChar(fFFDateMasks[0], fFFFieldsMax * Sizeof(TDateMaskRec), 0);
end;

procedure TWCSFlatFileDataset.ReadLayout;
var
  LayoutStream : TStream;
  FSLayout : TFileStream;
  NewFT : TFieldType;
  i, NewFS, NewFD, RollingOffs: integer;
  FD: TFieldDef;
  MaxBuff: integer;
  FDName: String;
begin
  if Assigned(fFFOnGetLayoutStream) then
  begin
    fFFOnGetLayoutStream(Self, LayoutStream, False);
    LayoutStream.Position := 0;
    FSLayout := Nil;
  end else begin
    if (ffLayoutFile = '') then raise exception.Create('No Layout Specified');
    FSLayout := TFileStream.Create(fFFLayoutFile, fmOpenReadWrite or fmShareDenyWrite);
    LayoutStream := FSLayout;
  end;
  try
    fDefinitions.StreamIn(LayoutStream);
  finally
    if FSLayout <> Nil then
      FSLayout.Free;
  end;

  AllocLayoutBuffers(fDefinitions.Count+2);
  if (fDefinitions.RecLen = 0) then
    raise exception.Create('No record length assigned');

  FieldDefs.Clear;
  RollingOffs := fDefinitions.RecHead;
  // dealing with 1 based arrays now for fields - skip auto filler
  with fFFRecBuffInfo do
  begin
    for i := 0 to fDefinitions.Count - 2 do
    with fDefinitions[i] do
    begin
      NewFS := 0;
      fFFFieldOffs[i+1] := RollingOffs;
      FieldDefinitionToDelphi(fDefinitions[i], NewFT, NewFS, NewFD, fFFDateMasks[i+1]);
      FDName := DName;
      if FieldDefs.IndexOf(FDName) >= 0 then
        FDName := IncrementalName(FDName,FieldDefExistsProc, Self, 2);
      FieldDefs.Add(FDName,NewFT,NewFS,False);
      FD := FieldDefs[FieldDefs.Count-1];
      case FD.DataType of
        ftBCD: FD.Precision := NewFS;
      end;
      fFFRecBuffInfo.fFFFieldLens[i+1] := DSrcSize;
      fFFCmpBuffInfo.fFFFieldLens[i+1] := DSrcSize;
      inc(RollingOffs, DSrcSize);
    end;
    i := fDefinitions.Count;
    FieldDefs.Add('<OriginalRecNo>',FTInteger,0,False);
    fFFFieldOffs[i] := -1;
    fFFFieldOffs[i] := -1;
    FieldDefs.Add('<FilteredRecNo>',FTInteger,0,False);
    inc(i);
    fFFFieldOffs[i] := -2;
    fFFFieldOffs[i] := -2;
    fFFCmpBuffInfo.fFFFieldCount := fFFFieldCount;
    for i := 0 to fFFFieldCount - 1 do
    begin
      fFFCmpBuffInfo.fFFFieldOffs[i] := fFFFieldOffs[i];
      fFFCmpBuffInfo.fFFFieldLens[i] := fFFFieldLens[i];
    end;
  end;

  CheckIndexFields(fFFIndexFields);
end;

procedure TWCSFlatFileDataset.FreeData;
var
  i : integer;
begin
  if fFFFreeStream then
  begin
    fFFStream.Free;
    fFFFreeStream := False;
  end;
  // inherited will clear the list - just free the objects
  for i := 0 to fFFRecList.Count-1 do
    Dispose(TWCSFlatFileRecord(fFFRecList[i]));
  for i := 0 to fFFHeadList.Count-1 do
    Dispose(TWCSFlatFileRecord(fFFHeadList[i]));
  ReAllocMem(fFFRecBuffInfo.fFFRecBuff, 0);
  ReAllocMem(fFFCmpBuffInfo.fFFRecBuff, 0);
  ReAllocWCSFFTransBuff(fFFRecBuffInfo.fFFFldBuff, 0);
  ReAllocWCSFFTransBuff(fFFCmpBuffInfo.fFFFldBuff, 0);
  fFFRecBuffInfo.fFFRecBuff := Nil;
  fFFCmpBuffInfo.fFFRecBuff := Nil;
  fFFRecList.Clear;
  fFFHeadList.Clear;
  fFFStream := Nil;
end;

function TWCSFlatFileDataset.FlushIn: Boolean;
begin
  if fFFReadBuffLeft > 0 then
  begin
    move(fFFReadBuff[fFFReadBuffPosn], fFFReadBuff[0], fFFReadBuffLeft);
    fFFReadBuffLeft := fFFReadBuffLeft +
      fFFStream.Read(fFFReadBuff[fFFReadBuffLeft], cWCSFFReadBuffSize-fFFReadBuffLeft);
  end else
    fFFReadBuffLeft := fFFStream.Read(fFFReadBuff[0], cWCSFFReadBuffSize);
  fFFReadBuffPosn := 0;
  Result := fFFReadBuffLeft > 0;
end;

function TWCSFlatFileDataset.FlushOut: Boolean;
begin
  if fFFWriteBuffPosn > 0 then
    Result := fFFStream.Write(fFFWriteBuff[0], fFFWriteBuffPosn) = fFFWriteBuffPosn
  else
    Result := True;
  fFFWriteBuffPosn := 0;
  fFFWriteBuffLeft := cWCSFFReadBuffSize;
end;

function TWCSFlatFileDataset.ReadStreamBuff(var Buff; const BuffLen: Longint): Longint;
var
  BuffNeeded : Longint;
begin
  Result := 0;
  while Result < BuffLen do
  begin
    if fFFReadBuffLeft = 0 then
      if not FlushIn then Exit; // not enough...
    if fFFReadBuffLeft > BuffLen-Result then
      BuffNeeded := BuffLen-Result
    else
      BuffNeeded := fFFReadBuffLeft;
    Move(fFFReadBuff[fFFReadBuffPosn], PChar(@Buff)[Result], BuffNeeded);
    inc(fFFReadBuffPosn, BuffNeeded);
    dec(fFFReadBuffLeft, BuffNeeded);
    inc(Result, BuffNeeded);
  end;
end;

function TWCSFlatFileDataset.WriteStreamBuff(const Buff; const BuffLen: Longint): Longint;
var
  BuffNeeded : Longint;
begin
  Result := 0;
  while Result < BuffLen do
  begin
    if fFFWriteBuffLeft = 0 then FlushOut;
    if fFFWriteBuffLeft > BuffLen-Result then
      BuffNeeded := BuffLen-Result
    else
      BuffNeeded := fFFWriteBuffLeft;
    Move(PChar(@Buff)[Result], fFFWriteBuff[fFFWriteBuffPosn], BuffNeeded);
    inc(fFFWriteBuffPosn, BuffNeeded);
    dec(fFFWriteBuffLeft, BuffNeeded);
    inc(Result, BuffNeeded);
  end;
end;

procedure TWCSFlatFileDataset.ReadData;
var
  RollingPos, RollingSize, SavePosn: Int64;
  TmpRec, ARec: TWCSFlatFileRecord;
  FilterNeeded, FilterIn : Boolean;
  i,FltCount : integer;
  DLRecLen, DLTotLen: Longint;
begin
  if fIndexError then  raise exception.Create('Index Error');
  try
    if Assigned(fFFOnGetDataStream) then
    begin
      fFFOnGetDataStream(Self, fFFStream, False);
      fFFFreeStream := False;
    end else begin
      fFFStream := TFileStream.Create(fFFDataFile, fmOpenRead or fmShareDenyWrite);
      fFFFreeStream := True;
    end;
    case fFFRecType of
      wffl_Delimited : RollingPos := 0;
//!!!!VarLenCheck
      else RollingPos    := fDefinitions.Header;
    end;
    fFFRecLen     := fDefinitions.RecLen;
    fFFRecOffs    := fDefinitions.RecHead;
    fFFRecDataLen := fFFRecLen - fFFRecOffs - fDefinitions.RecFoot;
    RollingSize := fFFStream.Size;
    fFFStream.Position := RollingPos;
    GetMem(fFFRecBuffInfo.fFFRecBuff, fFFRecLen);
    GetMem(fFFCmpBuffInfo.fFFRecBuff, fFFRecLen);
    ReAllocWCSFFTransBuff(fFFRecBuffInfo.fFFFldBuff, fFFRecLen);
    ReAllocWCSFFTransBuff(fFFCmpBuffInfo.fFFFldBuff, fFFRecLen);
    FilterNeeded := Assigned(fFFOnPreFilter);
    fTotalRecCount := 0;
    FltCount := 0;
// Move past parsed or delimited headers...
    case fFFRecType of
      wffl_Delimited :
        begin
          // Read the number of LINES of header
          RollingPos := 0;
          for i := 0 to fDefinitions.HeaderLines - 1 do
          begin
            if ReadDelimitedLine(DLRecLen, DLTotLen) then
            begin
              if DLRecLen > 0 then
              begin
                if DLRecLen > fFFRecLen then
                begin
                  fFFRecLen := DLRecLen;
                  ReAllocMem(fFFRecBuffInfo.fFFRecBuff, fFFRecLen);
                end;
                ARec := WCSFFNewRec(RollingPos, i+1, i+1, DLRecLen);
                fFFHeadList.Add(ARec);
                inc(RollingPos,DLTotLen);
              end;
            end else
              EXIT;
          end;
          while ReadDelimitedLine(DLRecLen, DLTotLen) do
          begin
            if DLRecLen > 0 then
            begin
              FilterIn := True;
              if DLRecLen > fFFRecLen then
              begin
                fFFRecLen := DLRecLen;
                ReAllocMem(fFFRecBuffInfo.fFFRecBuff, fFFRecLen);
              end;
              ARec := Nil;
              try
                if FilterNeeded then
                begin
                  if (fFFReadBuffPosn >= DLRecLen) then
                    move(fFFReadBuff[fFFReadBuffPosn-DLRecLen], fFFRecBuffInfo.fFFRecBuff[0], DLRecLen)
                  else begin
                    SavePosn := fFFStream.Position;
                    fFFStream.Position := RollingPos;
                    fFFStream.Read(fFFRecBuffInfo.fFFRecBuff[0], DLRecLen);
                    fFFStream.Position := SavePosn;
                  end;
                  ARec := WCSFFNewRec(RollingPos, fTotalRecCount, FltCount, DLRecLen);
                  fFFRecBuffInfo.fFFBuffRec := ARec;
                  ReadDelimitedFields(@fFFRecBuffInfo, fFFDelimQuotes[wffqOpenData],
                                      fFFDelimQuotes[wffqCloseData]);
                  fFFOnPreFilter(@fFFRecBuffInfo, FilterIn);
                end;
                if FilterIn then
                begin
                  if ARec = Nil then ARec := WCSFFNewRec(RollingPos, fTotalRecCount, FltCount, DLRecLen);
                  fFFRecList.Add(ARec);
                  ARec := Nil;
                  inc(FltCount);
                end else
              finally
                if ARec <> Nil then Dispose(ARec);
              end;
              inc(fTotalRecCount);
            end;
            inc(RollingPos,DLTotLen);
          end;
          if fFFRecLen > fDefinitions.RecLen then
          begin
            ReAllocWCSFFTransBuff(fFFRecBuffInfo.fFFFldBuff, fFFRecLen);
            ReAllocWCSFFTransBuff(fFFCmpBuffInfo.fFFFldBuff, fFFRecLen);
          end;
          EXIT;
        end;
      wffl_Parsed   :
        begin
          raise exception.Create('Parsed Text records not currently supported');
        end;
      wffl_Variable :
        begin
          raise exception.Create('Variable Length records not currently supported');
        end;
      wffl_Fixed :
        begin
        // standard Fixed length...
      //!!!!VarLenCheck
          while RollingPos+fFFRecLen <= RollingSize do
          begin
      //!!!!VarLenCheck
            if ReadStreamBuff(fFFRecBuffInfo.fFFRecBuff^, fFFRecLen) < fFFRecLen then Break;
            ARec := WCSFFNewRec(RollingPos, fTotalRecCount, FltCount, fFFRecLen);
            try
              fFFRecBuffInfo.fFFBuffRec := ARec;
              FilterIn := True;
              if FilterNeeded then fFFOnPreFilter(@fFFRecBuffInfo, FilterIn);
              if FilterIn then
              begin
        //!!!!VarLenCheck
        //!!!! Here - if we want "All in Memory" - copy it and adjust Offset !!!!
                fFFRecList.Add(ARec);
                ARec := Nil;
                inc(FltCount);
              end else
            finally
              if ARec <> Nil then Dispose(ARec);
            end;
      //!!!!VarLenCheck
            inc(RollingPos, fFFRecLen);
            inc(fTotalRecCount);
          end;
        end;
    end;
  except
    FreeData;
    raise;
  end;
end;

procedure TWCSFlatFileDataset.InternalOpen;
begin
  try
    inherited;
  except
    FreeData;
    raise;
  end;
end;

procedure TWCSFlatFileDataset.InternalClose;
begin
  fTotalRecCount := 0;
  if fFFStream = Nil then exit;
  try
    inherited;
  finally
    FreeData;
  end;
end;

procedure TWCSFlatFileDataset.ListSetupFieldDefs;
begin
  ReadLayout;
end;

function TWCSFlatFileDataset.ListCreateElement(const Element: TObject): TObject;
begin
  raise exception.Create(SDataSetReadOnly);
end;

procedure TWCSFlatFileDataset.ListFreeElement(const Element: TObject);
begin
  raise exception.Create(SDataSetReadOnly);
//  Dispose(TWCSFlatFileRecord(fFFRecList[RNo]));
end;

procedure TWCSFlatFileDataset.ListInsert(const NewObj : TObject; var NewIdx: integer);
begin
  raise exception.Create(SDataSetReadOnly);
end;

procedure TWCSFlatFileDataset.ListDelete(const DelIdx: integer);
begin
  raise exception.Create(SDataSetReadOnly);
end;

procedure TWCSFlatFileDataset.ListSetObject(NewObj: TObject;
  var NewIdx: Integer);
begin
  raise exception.Create(SDataSetReadOnly);
end;

// DelphiVar will either be fFFFldBuf^ or fFFFldCmpBuf^
procedure TWCSFlatFileDataset.DelphivarFromField(
  const RecBuffInfo: TFFBufferInfo;
  const FldNo: integer);
var
  SrcBuff      : PFFRecBuff;
  Def          : TFieldDefinition;
  TransStr     : String;
  TransMax     : Longint;
  si, E: integer;
  B, BP: Byte;
  WorkDate: TDateTime;
begin
  Def := fDefinitions[FldNo-1];
//!!!! This must deal with Mask translation!
  with RecBuffInfo^, fFFFldBuff do
  begin

  // 1) Get value into fFFFldBuff in Source type format
    case fFFRecType of
      wffl_Delimited:
        begin
        //    This will result in a String only
          TransUsage := wffu_Chars;
          TransLen := ReadDelimitedField(RecBuffInfo, FldNo,
                                          fFFDelimQuotes[wffqOpenData], fFFDelimQuotes[wffqCloseData]);
          TransNull := TransLen = 0;


        end;
      wffl_Fixed:
        //    This will result in a String, INTEGER, LARGINT or DOUBLE only
        begin
          SrcBuff := @fFFRecBuff^[fFFFieldOffs[FldNo]];
          TransNull := False;
          TransLen := Def.DSrcSize;
          case Def.DSrcType of
            wffs_Chars    :
              begin
                TransUsage := wffu_Chars;
                // Trailing Space-strip source....
                while (TransLen > 0) and (SrcBuff[TransLen-1] in [0,32]) do
                  dec(TransLen);
                TransPC[TransLen] := #0;
                if TransLen > 0 then
                  Move(SrcBuff^, TransPC[0], TransLen)
                else
                  TransNull := True;
              end;
            wffs_NumPk02  :// Packed Numeric (base 2 storage)
              begin
                // REALLY BIG integers just end up as Floats.....
                // Signed?
                TransFloat^ := 0;
                TransSigned := (Def.DSrcSignStorage <> nfNone) and (SrcBuff^[0] and $80 > 0);
                if TransSigned then
                  TransFloat^ := (SrcBuff^[0] and $7F)
                else
                  TransFloat^ := SrcBuff^[0];
                for si := 1 to TransLen - 1 do
                  TransFloat^ := (TransFloat^ * 256) + SrcBuff^[si];
                if TransSigned then
                  TransFloat^ := - TransFloat^;
                if (Def.DSrcDP > 0) then TransFloat^ := TransFloat^ / (10 * Def.DSrcDP);
                if (Def.DSrcDP = 0) and (TransLen <= 8) then
                begin
                  if TransLen <= 4 then
                  begin
                    TransUsage := wffu_LongInt;
                    TransLong^ := Trunc(TransFloat^);
                  end else begin
                    TransUsage := wffu_LargeInt;
                    TransInt64^ := Trunc(TransFloat^);
                  end;
                end else
                  TransUsage := wffu_Float;
              end;
            wffs_NumPk10 :  // Packed Numeric (Nibble per digit base 10 storage)
              begin
              // These are stored High to Low with signs...
              // MSB has partial, LSB has Signed (even if not "signed" in picture)
                TransLen := Def.DSrcPrec;
                TransFloat^ := 0;
                // 1234 => 01 23 4S  : Prec = 5
                for si := 0 to TransLen - 1 do
                begin
                  BP := si div 2;
                  if Odd(si) then
                    B := ord(SrcBuff[BP]) and $F
                  else
                    B := (ord(SrcBuff[BP]) shr 4) and $F;
                  TransFloat^ := (TransFloat^ * 10) + B;
                end;
                TransSigned := (Def.DSrcSignStorage <> nfNone) and
                               (ord(SrcBuff[TransLen-1]) and $F = $D);
                if (Def.DSrcDP > 0) then TransFloat^ := TransFloat^ / (10 * Def.DSrcDP);
                if TransSigned then TransFloat^ := 0.0 - TransFloat^;
                if (Def.DSrcDP = 0) and (TransLen <= 9) then
                begin
                  if TransLen <= 4 then
                  begin
                    TransUsage := wffu_LongInt;
                    TransLong^ := Trunc(TransFloat^);
                  end else begin
                    TransUsage := wffu_LargeInt;
                    TransInt64^ := Trunc(TransFloat^);
                  end;
                end else
                  TransUsage := wffu_Float;
              end;
            wffs_NumUnPk :  // UnPacked Numeric (byte per digit base 10 storage)
              begin
        // DEAL WITH SEPARATE/LEADING/TRAILING DIGH HERE
        // CURRENTLY ASSUMING TRAILING?
                TransFloat^ := 0;
                TransLen := Def.DSrcPrec;
                for si := 0 to TransLen - 1 do
                begin
                  if SrcBuff[si] >= 48 then
                    B := (SrcBuff[si] and $3F) - 48
                  else
                    B := 0;
                  TransFloat^ := (TransFloat^ * 10) + B;
                end;
                TransSigned := (Def.DSrcSignStorage <> nfNone) and (SrcBuff[TransLen-1] - B = $70);
                if TransSigned then
                  TransFloat^ := 0.0 - TransFloat^;
                if (Def.DSrcDP > 0) then TransFloat^ := TransFloat^ / (10 * Def.DSrcDP);
                if (Def.DSrcDP = 0) and (TransLen <= 18) then
                begin
                  if TransLen <= 9 then
                  begin
                    TransUsage := wffu_LongInt;
                    TransLong^ := Trunc(TransFloat^);
                  end else begin
                    TransUsage := wffu_LargeInt;
                    TransInt64^ := Trunc(TransFloat^);
                  end;
                end else
                  TransUsage := wffu_Float;
              end;
            wffs_Single :   // Ansi/ieee single precision float
              begin
                TransUsage := wffu_Float;
                TransLen := 8;
                TransFloat^ := PSingle(SrcBuff)^;
              end;
            wffs_Double :   // Ansi/ieee double precision float
              begin
                TransUsage := wffu_Float;
                TransLen := 8;
                TransFloat^ := PDouble(SrcBuff)^;
              end;
            wffs_Binary : // All other data keep in string for now...
              begin
                TransUsage := wffu_Binary;
                TransLen   := Def.DSrcSize;
                Move(SrcBuff^, TransPC[0], Def.DSrcSize);
              end;
          end;
        end;
    end;

    // Phase 2) Translate using mask or standard Delphi conversions
    //          NOTE: at this stage there cannot be any DATE types (yet!)
    if (not TransNull) and (TransUsage <> Def.DUsgType) then
    begin
      case Def.DUsgType of
        wffu_Chars :
        begin
          case TransUsage of
            wffu_LongInt : // 1-4 byte Integer.
              begin
                TransStr := IntToStr(TransLong^);
              end;
            wffu_LargeInt: // 5-8 byte Integer.
              begin
                TransStr := IntToStr(TransInt64^);
              end;
            wffu_Float   :
              begin
                TransStr := FloatToStr(TransFloat^);
              end;
            wffu_BCD     :
              begin
                TransStr := FloatToStr(BCDToDouble(TransBCD^));
              end;
          end;
          if TransUsage <> wffu_Binary then
          begin
            TransLen := Length(TransStr);
            StrPCopy(TransPC, TransStr);
          end;
        end;
  //!!!! Lots to fill in here!!!!
        wffu_LongInt  :// 1-4 byte Integer.
          case TransUsage of
            wffu_Chars :
              begin
                TransNull := not TryStrToInt(StrPas(TransPC), TransLong^);
              end;
            wffu_LargeInt: // 5-8 byte Integer.
              begin
                TransLong^ := TransInt64^;
              end;
            wffu_Float   : // Floating point.
              begin
                TransLong^ := Trunc(TransFloat^);
              end;
            wffu_BCD     : // Large numerics that will not fit in Integer or Float.
              begin
                TransLong^ := BcdToInteger(TransBCD^, True);
              end;
          end;
        wffu_LargeInt : // 5-8 byte Integer.
          case TransUsage of
            wffu_Chars :
              begin
                TransNull := not TryStrToInt64(StrPas(TransPC), TransInt64^);
              end;
            wffu_LongInt : // 1-4 byte Integer.
               begin
                TransInt64^ := TransLong^;
               end;
            wffu_Float   : // Floating point.
               begin
                TransInt64^ := Trunc(TransFloat^);
               end;
            wffu_BCD     : // Large numerics that will not fit in Integer or Float.
              begin
                TransInt64^ := Trunc(BcdToDouble(TransBCD^));
              end;
          end;
        wffu_Float    : // Floating point.
          case TransUsage of
            wffu_Chars :
              begin
                TransNull := not TryStrToFloat(StrPas(TransPC), TransFloat^);
              end;
            wffu_LongInt : // 1-4 byte Integer.
              begin
                TransFloat^ := TransLong^;
              end;
            wffu_LargeInt: // 5-8 byte Integer.
              begin
                TransFloat^ := TransInt64^;
              end;
            wffu_BCD     : // Large numerics that will not fit in Integer or Float.
              begin
                TransFloat^ := BcdToDouble(TransBCD^);
              end;
          end;
        wffu_BCD      : // Large numerics that will not fit in Integer or Float.
          case TransUsage of
            wffu_Chars :
              begin
                TransNull := not TryStrToBcd(StrPas(TransPC), TransBCD^);
              end;
            wffu_LongInt : // 1-4 byte Integer.
              begin
                TransBCD^ := StrToBcd(IntToStr(TransLong^));
              end;
            wffu_LargeInt: // 5-8 byte Integer.
              begin
                TransBCD^ := StrToBcd(IntToStr(TransInt64^));
              end;
            wffu_Float   : // Floating point.
               begin
                 TransBCD^ := StrToBcd(FloatToStr(TransFloat^));
               end;
          end;
        wffu_DateTime : // All dates and times.
          begin
            case TransUsage of
              wffu_Chars :
                begin
                  SetString(TransStr, TransPC, TransLen);
                end;
              wffu_LongInt : // 1-4 byte Integer.
                begin
                  TransStr := IntToStr(TransLong^);
                end;
              wffu_LargeInt: // 5-8 byte Integer.
                begin
                  TransStr := IntToStr(TransInt64^);
                end;
              wffu_Float   :
                begin
                  TransStr := FloatToStr(TransFloat^);
                end;
              wffu_BCD     :
                begin
                  TransStr := FloatToStr(BCDToDouble(TransBCD^));
                end;
              wffu_Binary  : // Anything else!
                begin
                  SetString(TransStr, TransPC, TransLen);
                end;
            end;
            TransNull := not ExtractMaskedDate(TransStr, fFFDateMasks[FldNo], TransDate^);
          end;
        wffu_Binary :
          begin
            TransNull := True;
            for si := 0 to TransLen-1 do
              if TransPC[si] <> #0 then
              begin
                TransNull := False;
                Break;
              end;
          end;
      end; {Case Def.DUsgType}
      TransUsage := Def.DUsgType;
    end; {if DelphiVarType <> Def.DUsgType}
  end;
end;


// Getting field data is a 3 phase operation :
// 1) Get the data out of the Stream into a Delphi var.
// 2) Use Mask and usage type to translate simple Delphi var.
// These 2 phases also used in indexing, so need separating (DelphivarFromField).
// 3) Use final field type to output to field buffer (below).
function TWCSFlatFileDataset.ListGetField(ListObject : TObject; ListIndex: integer;
  Field: TField; FieldBuf: Pointer): Boolean;
type
  PDateTimeRec = ^TDateTimeRec;
var
  Def : TFieldDefinition;
  FldNo, si: integer;
  TransMax    : Longint;
  TimeStamp: TTimeStamp;

  B : Byte;
  BP :integer;
begin
  FldNo := Field.FieldNo;
  Result := False;
  if (ListObject <> Nil) and (FldNo > 0) then
  with fFFRecBuffInfo do
  begin
    if fFFFieldOffs[FldNo] >= 0 then
    with fFFFldBuff do
    begin
      if FieldBuf <> Nil then FillChar(FieldBuf^, Field.DataSize, 0);
      Def := fDefinitions[FldNo-1];
      case fFFRecType of
        wffl_Delimited:
          begin
            //if FldNo > fFFFieldCount then Exit;
            if fFFBuffRec <> TWCSFlatFileRecord(ListObject) then
            begin
              fFFBuffRec := TWCSFlatFileRecord(ListObject);
              fFFStream.Position := fFFBuffRec.fOffset;
              ReadDelimitedFields(@fFFRecBuffInfo,fFFDelimQuotes[wffqOpenData],
                                  fFFDelimQuotes[wffqCloseData]);
            end;
            if fFFFieldLens[FldNo] < 1 then Exit;
          end;
        wffl_Fixed:
          begin
            fFFStream.Position := TWCSFlatFileRecord(ListObject).fOffset + fFFFieldOffs[FldNo];
            fFFStream.Read(fFFRecBuff^[fFFFieldOffs[FldNo]], fFFFieldLens[FldNo]);
          end;
      end;
      DelphivarFromField(@fFFRecBuffInfo, FldNo);
      Result := not fFFFldBuff.TransNull;
      if (FieldBuf <> Nil) and Result then
      begin
        TransMax := Field.DataSize;
        // Phase 3) Translate to Delphi Field Buffer
        case Def.DUsgType of
          wffu_Chars :
            begin
              if TransLen > TransMax then
                Move(TransPC[0], FieldBuf^, TransMax)
              else
                Move(TransPC[0], FieldBuf^, TransLen);
            end;
          wffu_LongInt : // 1-4 byte Integer.
             begin
               PInteger(FieldBuf)^ := TransLong^;
             end;
          wffu_LargeInt: // 5-8 byte Integer.
             begin
               PInt64(FieldBuf)^ := TransInt64^;
             end;
          wffu_Float   : // Floating point.
             begin
               PDouble(FieldBuf)^ := TransFloat^;
             end;
          wffu_BCD     : // Large numerics that will not fit in Integer or Float.
             begin
               DoubleToBcd(TransFloat^, PBCD(FieldBuf)^);
             end;
          wffu_DateTime: // All dates and times.
             begin
               try
                 TimeStamp := DateTimeToTimeStamp(TransDate^);
                 case Field.DataType of
                  ftDate: PDateTimeRec(FieldBuf)^.Date := TimeStamp.Date;
                  ftTime: PDateTimeRec(FieldBuf)^.Time := TimeStamp.Time;
                else
                  PDateTimeRec(FieldBuf)^.DateTime := TimeStampToMSecs(TimeStamp);
                  end;
               except
                 Result := False;
                 PDateTimeRec(FieldBuf)^.DateTime := 1;
               end;
             end;
          wffu_Binary  : // Anything else!
             begin
               Move(TransPC[0], FieldBuf^, Def.DSrcSize);
             end;
        end;
        if Field.DataType = ftString then
          Pchar(FieldBuf)[TransMax] := #0;
      end;
    end else
      case fFFFieldOffs[FldNo] of
        -1 : begin
               if (FieldBuf <> Nil) then
                 PInteger(FieldBuf)^ := TWCSFlatFileRecord(ListObject).fRecNo;
               Result := True;
             end;
        -2 : begin
               if (FieldBuf <> Nil) then
                 PInteger(FieldBuf)^ := TWCSFlatFileRecord(ListObject).fFltNo;
               Result := True;
             end;
      end;
  end;
end;

procedure TWCSFlatFileDataset.ListSetField(ListObject: TObject;
  Field: TField; FieldBuf: Pointer);
begin
  raise exception.Create(SDataSetReadOnly);
end;

procedure TWCSFlatFileDataset.ListSetupFields;
var
  i, j : integer;
  Def          : TFieldDefinition;
begin
// try to set some default masks
  if DefaultFields then
  begin
    for i := 0 to FieldCount - 3 do
    if (Fields[i].FieldNo > 0) and (Fields[i] is TNumericField) then
    with TNumericField(Fields[i]) do
    begin
      Def := fDefinitions[FieldNo-1];
      if Def.DSrcDP > 0 then
      begin
        DisplayFormat := '0.0';
        for j := 2 to Def.DSrcDP do
          DisplayFormat := DisplayFormat + '0';
      end;
    end;
  end;
end;

procedure TWCSFlatFileDataset.ListDataFetch;
begin
  ReadData;
end;

function TWCSFlatFileDataset.IndexCompare(Object1, Object2: TObject): Integer;

  procedure CompareInts(const I1, I2: integer);
  begin
    if I1 < I2 then Result := -1 else
    if I1 > I2 then Result := 1;
  end;

  procedure CompareLargeInts(const I1, I2: int64);
  begin
    if I1 < I2 then Result := -1 else
    if I1 > I2 then Result := 1;
  end;

  procedure CompareFloats(const I1, I2: Extended);
  begin
    if I1 < I2 then Result := -1 else
    if I1 > I2 then Result := 1;
  end;

  procedure CompareDateTimes(const I1, I2: TDatetime);
  begin
    if I1 < I2 then Result := -1 else
    if I1 > I2 then Result := 1;
  end;

  procedure CompareStrings(const I1, I2: PChar; const Len1, Len2: integer);
  var
    i, Len : integer;
  begin
    i := 0;
    if Len1 < Len2 then Len := Len1 else Len := Len2;
    while (i < Len) and (Result = 0) do
      if Upcase(I1[i]) < Upcase(I2[i]) then Result := -1 else
      if Upcase(I1[i]) > Upcase(I2[i]) then Result := 1 else
      if I1[i] < I2[i] then Result := -1 else
      if I1[i] > I2[i] then Result := 1 else
        inc(i);
    if (Result = 0) then CompareInts(Len1, Len2);
  end;

  procedure CompareBins(const I1, I2: PChar; const Len1, Len2: integer);
  var
    i, Len : integer;
  begin
    i := 0;
    if Len1 < Len2 then Len := Len1 else Len := Len2;
    while (i < Len) and (Result = 0) do
      if I1[i] < I2[i] then Result := -1 else
      if I1[i] > I2[i] then Result := 1 else
        inc(i);
    if (Result = 0) then CompareInts(Len1, Len2);
  end;


var
  i : integer;
  Def          : TFieldDefinition;
begin
  Result := 0;
  if fIndexFieldCount > 0 then
  with fFFRecBuffInfo.fFFFldBuff do
  begin
//!!!!VarLenCheck
    fFFRecBuffInfo.fFFBuffRec := TWCSFlatFileRecord(Object1);
    fFFCmpBuffInfo.fFFBuffRec := TWCSFlatFileRecord(Object2);
    fFFStream.Position := TWCSFlatFileRecord(Object1).fOffset;
    case fFFRecType of
      wffl_Delimited:
        begin
          ReadDelimitedFields(@fFFRecBuffInfo,
                              fFFDelimQuotes[wffqOpenData],
                              fFFDelimQuotes[wffqCloseData]);
        end;                              
      wffl_Fixed: fFFStream.Read(fFFRecBuffInfo.fFFRecBuff^[0], fffRecLen);
    end;
    fFFStream.Position := TWCSFlatFileRecord(Object2).fOffset;
    case fFFRecType of
      wffl_Delimited: ReadDelimitedFields(@fFFCmpBuffInfo,
                                          fFFDelimQuotes[wffqOpenData],
                                          fFFDelimQuotes[wffqCloseData]);
      wffl_Fixed: fFFStream.Read(fFFCmpBuffInfo.fFFRecBuff^[0], fffRecLen);
    end;
    i := 0;
    while (Result = 0) and (i < fIndexFieldCount) do
    begin
      Def := fDefinitions[fIndexFieldNos[i]-1];
      DelphivarFromField(@fFFRecBuffInfo, fIndexFieldNos[i]);
      DelphivarFromField(@fFFCmpBuffInfo, fIndexFieldNos[i]);
      case Def.DUsgType of
        wffu_Chars     : CompareStrings(TransPC, fFFCmpBuffInfo.fFFFldBuff.TransPC, TransLen, fFFCmpBuffInfo.fFFFldBuff.TransLen);
        wffu_LongInt   : CompareInts(TransLong^, fFFCmpBuffInfo.fFFFldBuff.TransLong^);
        wffu_LargeInt  : CompareLargeInts(TransInt64^, fFFCmpBuffInfo.fFFFldBuff.TransInt64^);
        wffu_Float     : CompareFloats(TransFloat^, fFFCmpBuffInfo.fFFFldBuff.TransFloat^);
//!!!! These are stored by Delphi as TCurrency in a TBCDfield - USELESS!!!!
// This gives us NO scope for the potentially really big numbers (i.e. 32 dp precision)!
//!!!! Must check dbPAL datasets etc - are they assumuing storage as proper BCD?
//!!!! ftFmtBCD must now be used !!!!
        wffu_BCD       : BcdCompare(TransBCD^, fFFCmpBuffInfo.fFFFldBuff.TransBCD^);
        wffu_DateTime  : CompareDateTimes(TransDate^, fFFCmpBuffInfo.fFFFldBuff.TransDate^);
        else             CompareBins(TransPC, fFFCmpBuffInfo.fFFFldBuff.TransPC, TransLen, fFFCmpBuffInfo.fFFFldBuff.TransLen);
      end;
      inc(i);
    end;
  end;
  if Result = 0 then
    CompareLargeInts(TWCSFlatFileRecord(Object1).fOffset, TWCSFlatFileRecord(Object2).fOffset);
end;

procedure TWCSFlatFileDataset.SetFFIndexFields(const Value: String);
begin
  if fFFIndexFields <> Value then
  begin
    if Active then
    begin
      CheckIndexFields(Value);
      if fIndexError then
      begin
        fIndexError := False;
        raise exception.Create('Index Error');
      end;
    end;
    fFFIndexFields := Value;
    if Active then
    begin
      SortByFunc(IndexCompare);
    end;
  end;

end;

procedure TWCSFlatFileDataset.ListDataRead;
begin
  if fIndexFieldCount > 0 then
    SortByFunc(IndexCompare);
end;

function TWCSFlatFileDataset.ReadDelimitedLine(var RecLen, TotLen: Longint): Boolean;
var
  i, j : integer;
begin
 // use the buffer directly and return the TOTAL Length Read and that not including delim..
 RecLen := 0;
 TotLen := 0;
 while True do
 begin
    if fFFReadBuffLeft <= fFFRecDelimLen then
      if (not FlushIn) or (fFFReadBuffLeft <= fFFRecDelimLen) then
      begin
      // Deal with end of line situation
        inc(TotLen, fFFReadBuffLeft);
        if fFFRecDelimCR then
        begin
          while (fFFReadBuffLeft > 0) do
          begin
            if (not (fFFReadBuff[fFFReadBuffPosn] in [#13,#10])) then
              inc(RecLen);
            inc(fFFReadBuffPosn);
            dec(fFFReadBuffLeft);
          end;
        end else begin
          while (fFFReadBuffLeft > 0) do
          begin
            if (fFFReadBuff[fFFReadBuffPosn] <> fFFRecDelim[1]) then
              inc(RecLen);
            inc(fFFReadBuffPosn);
            dec(fFFReadBuffLeft);
          end;
        end;
        Break; // must be at end of file
      end;
    if fFFRecDelimCR then
    begin
   // Deal CR delim quirky files
      if fFFReadBuff[fFFReadBuffPosn] in [#13,#10] then
      begin
        while (fFFReadBuff[fFFReadBuffPosn] in [#13,#10]) do
        begin
          inc(fFFReadBuffPosn);
          dec(fFFReadBuffLeft);
          inc(TotLen);
          if fFFReadBuffLeft = 0 then
            if (not FlushIn) then Exit;
        end;
        Break;
      end;
    end else begin
    // here we have GREATER than delimlen left....
      i := 1;
      while (i <= fFFRecDelimLen) and (fFFReadBuff[fFFReadBuffPosn] = fFFRecDelim[i]) do
      begin
        inc(fFFReadBuffPosn);
        dec(fFFReadBuffLeft);
        inc(i);
        inc(TotLen);
      end;
      if i > fFFRecDelimLen then
        Break
      else // increment by any extras already picked off
        inc(RecLen, i-1);
    end;
    // increment for the single char
    inc(fFFReadBuffPosn);
    dec(fFFReadBuffLeft);
    inc(RecLen);
    inc(TotLen);
 end;
 Result := TotLen > 0;
end;

// Fill in the Offsets, Lengths and count. Returns length of longest.
function TWCSFlatFileDataset.ReadDelimitedFields(const ARecBuffInfo: TFFBufferInfo;
  const OQ, CQ: Char): Integer;
var
  RP, FS, FL : integer;
  InQuotes, IsDelim: Boolean;
  C : Char;

  procedure ProcessField;
  var
    J : integer;
  begin
    with ARecBuffINfo^ do
    begin
      inc(fFFFieldCount);
      if fFFFieldCount >= fFFFieldsMax then
      begin
        fFFFieldsMax := fFFFieldCount + 64;
        SetLength(fFFFieldOffs, fFFFieldsMax);
        SetLength(fFFFieldLens, fFFFieldsMax);
      end;
      if FS >= 0 then
      begin
        J := RP;
        while (J > FS) and (fFFRecBuff[J-1] = 32) do dec(J);
        fFFFieldOffs[fFFFieldCount] := FS;
        fFFFieldLens[fFFFieldCount] := J-FS;
        if fFFFieldLens[fFFFieldCount] > Result then
          Result := fFFFieldLens[fFFFieldCount];
      end else begin
        fFFFieldOffs[fFFFieldCount] := RP;
        fFFFieldLens[fFFFieldCount] := 0;
      end;
    end;
  end;

  procedure CheckLongDelim;
  var
    J : integer;
  begin
    J := 2;
    while IsDelim and (J <= fFFFldDelimLen) and (RP+J-1 < ARecBuffInfo^.fFFBuffRec^.fLength) do
      if fFFFldDelim[J] = Char(ARecBuffINfo^.fFFRecBuff[RP+J-1]) then
        inc(J)
      else
        IsDelim := False;
  end;

begin
  Result := 0;
  with ARecBuffInfo^, fFFBuffRec^ do
  begin
//!!!! HERE - Watch overrun when open live (as opposed to Scanning)
    fFFFieldCount := 0;
    FS := -1;
    if fLength > fFFFldDelimLen then
    begin
      fFFStream.Position := fOffset;
      fFFStream.Read(fFFRecBuff[0], fLength);
      RP := 0;
      InQuotes := False;
      while RP < fLength do
      begin
        C := Char(fFFRecBuff[RP]);
        if InQuotes then
        begin
          if C = CQ then
            if (OQ=CQ) and (RP < fLength-1) and (Char(fFFRecBuff[RP+1]) = CQ) then
              inc(RP)
            else
              InQuotes := False;
        end else begin
          if C = fFFFldDelim[1] then
          begin
            IsDelim := True;
            if (fFFFldDelimLen > 1) then CheckLongDelim;
          end else
            IsDelim := False;
          if IsDelim then
          begin
            ProcessField;
            FS := -1;
          end else
            if C <> ' ' then
            begin
              if FS < 0 then FS := RP;
              if C = OQ then InQuotes := True;
            end;
        end;
        inc(RP);
      end;
    end;
    ProcessField;
  end;

end;

function TWCSFlatFileDataset.ReadDelimitedField(const ARecBuffInfo: TFFBufferInfo;
  const FldNo: integer;
  const OQ, CQ: Char): integer;
var
  FS, FP, FL : integer;
  FldBuffIn, FldBuffOut : PChar;
  InQuotes: Boolean;
  C : Char;
begin
  with ARecBuffInfo^ do
  begin
    FldBuffOut := fFFFldBuff.TransPC;
    Result := fFFFieldLens[FldNo];
    if Result > 0 then
    begin
      FP := 0;
      FL := 0;
      FldBuffIn := @fFFRecBuff[fFFFieldOffs[FldNo]];
      InQuotes := False;
      while FP < Result do
      begin
        C := FldBuffIn[FP];
        if InQuotes then
        begin
          if C = CQ then
            if (OQ=CQ) and (FP < Result-1) and (FldBuffIn[FP+1] = CQ) then
            begin
              FldBuffOut[FL] := C;
              inc(FL);
            end else
              inQuotes := False
          else begin
            FldBuffOut[FL] := C;
            inc(FL);
          end;
        end else
          if C = OQ then
            inQuotes := True
          else begin
            FldBuffOut[FL] := C;
            inc(FL);
          end;
        inc(FP);
      end;
      Result := FL;
    end;
    fFFFldBuff.TransNull := Result = 0;
    FldBuffOut[Result] := #0;
  end;
end;

procedure TWCSFlatFileDataset.FldBuffBestType(const Len: integer;
  var StartType: TWCSFlatFileBestScanType);
var
  i, DP : integer;
  C, SC : Char;
  TestDate: TDateTime;
begin
  with fFFRecBuffInfo do
  if (Len > 0) and (StartType <> wffbs_Chars) then
  begin
    DP := 0;
    SC := fFFFldBuff.TransPC[0];
    if StartType = wffbs_Any then
    begin
      if (SC in ['0'..'9']) or ((SC = '-') and (Len > 1)) then
      begin
        StartType := wffbs_Integer;
        for i := 1 to Len - 1 do
        begin
          C := fFFFldBuff.TransPC[i];
          case StartType of
            wffbs_Chars : Break;
            wffbs_Integer :
                if (i > 9) or (not (C in ['0'..'9'])) then
                  if (C in ['0'..'9']) or ((C = '.') and ((i > 1) or (SC <> '-'))) then
                  begin
                    StartType := wffbs_Float;
                    DP := 1;
                  end else
                    if (Len >= 6) and (i < 3) and (SC in DatePartChars) and
                      TryStrToDateTime(fFFFldBuff.TransPC, TestDate) then
                    begin
                      StartType := wffbs_Date;
                    end else
                      StartType := wffbs_Chars;
            wffbs_Float :
                if (i < 18) and ((C in ['0'..'9']) or ((DP = 0) and (C = '.') and ((i > 1) or (SC <> '-')))) then
                begin
                  if C = '.' then DP := 1;
                end else
                  if (Len >= 6) and (SC in DatePartChars) and
                     (((i < 3) and (DP = 0)) or (i <= 6)) and
                    TryStrToDateTime(fFFFldBuff.TransPC, TestDate) then
                  begin
                    StartType := wffbs_Date
                  end else
                    StartType := wffbs_Chars;
            wffbs_Date : Break;
          end;
        end;
      end else
        if (Len >= 6) and (SC in DatePartChars) and
           TryStrToDateTime(fFFFldBuff.TransPC, TestDate) then
        begin
          StartType := wffbs_Date;
        end else
          StartType := wffbs_Chars;
    end else begin
      for i := 0 to Len - 1 do
      begin
        C := fFFFldBuff.TransPC[i];
        case StartType of
          wffbs_Chars : Break;
          wffbs_Integer :
              if not ((C in ['0'..'9']) or ((i = 0) and (C = '-'))) then
                if (i > 0) and (C = '.') then
                begin
                  StartType := wffbs_Float;
                  DP := 1;
                end else
                  StartType := wffbs_Chars;
          wffbs_Float :
              if not ((C in ['0'..'9']) or ((i = 0) and (C = '-')) or ((DP = 0) and (i > 0) and (C = '.'))) then
                StartType := wffbs_Chars
              else
                if C = '.' then DP := 1;
          wffbs_Date :
              if (Len >= 6) and (SC in DatePartChars) and
                TryStrToDateTime(fFFFldBuff.TransPC, TestDate) then
                Break
              else
                StartType := wffbs_Chars;
        end;
      end;
    end;
  end else
    StartType := wffbs_Chars;
end;

function TWCSFlatFileDataset.PreScanToLayout(const FldDefs: TFieldDefinitionList): integer;
var
  SaveDefs: TFieldDefinitionList;
  WasEditing: Boolean;
  FixedFieldCount: Boolean;
  MaxFieldLen : integer;
  fi, ri, DefsFound : integer;
  Bst: TWCSFlatFileBestScanType;
  ScanBest: array of TWCSFlatFileBestScanType;
  ScanLen : array of integer;
begin
  Result := 0;
  CheckInactive;
  if FldDefs.Family <> ffdf_Delimited then
    raise exception.Create('Prescan only supported for delimited data');
  SaveDefs:= fDefinitions;
  try
    fDefinitions := FldDefs;
    WasEditing := FldDefs.Editing;
    if not WasEditing then  FldDefs.Edit;
    try
      FldDefs.Clear(False);
      DefsFound := 0;
      // make plenty of space for the parsed defs
      if FldDefs.RecLen = 0 then FldDefs.RecLen := 1024;
      try
        AllocLayoutBuffers(1024);
       // Prescan line ends and count....
        with fFFRecBuffInfo do
        try
          // This will extend RecBuff and FieldBuffs to longest record
          ReadData;
          // Delimiter scan header lines... THIS WILL DICTATE THE FIELDS - EXTRAS DROPPED!
          FixedFieldCount := False;
          if fFFHeadList.Count > 0 then
          begin
            fFFRecBuffInfo.fFFBuffRec := TWCSFlatFileRecord(fFFHeadList[0]);
            MaxFieldLen := ReadDelimitedFields(@fFFRecBuffInfo,
                                               fFFDelimQuotes[wffqOpenName],
                                               fFFDelimQuotes[wffqCloseName]);
            if fFFFieldCount > 0 then
            begin
              FixedFieldCount := True;
              DefsFound := fFFFieldCount;
              SetLength(ScanBest, DefsFound+1);
              SetLength(ScanLen, DefsFound+1);
              for fi := 0 to DefsFound - 1 do
              begin
                ScanBest[fi+1] := wffbs_Any;
                ScanLen[fi+1] := -1;
                FldDefs.AddDef(Format('FIELD%4.4d',[fi+1])).EditNoModify;
                if ReadDelimitedField(@fFFRecBuffInfo, fi+1,fFFDelimQuotes[wffqOpenName], fFFDelimQuotes[wffqCloseName]) > 0 then
                  FldDefs[fi].DName := fFFFldBuff.TransPC;
              end;
            end;
          end;
          // Now delimiter scan each line to get max Field count, max size, potential common types etc.
          Result := fFFRecList.Count;
          for ri := 0 to fFFRecList.Count - 1 do
          begin
            fFFRecBuffInfo.fFFBuffRec := TWCSFlatFileRecord(fFFRecList[ri]);
            MaxFieldLen := ReadDelimitedFields(@fFFRecBuffInfo,
                                               fFFDelimQuotes[wffqOpenData],
                                               fFFDelimQuotes[wffqCloseData]);
            if fFFFieldCount > DefsFound then
              if (not FixedFieldCount) then
              begin
                if DefsFound < fFFFieldCount then
                begin
                  SetLength(ScanBest, fFFFieldCount+1);
                  SetLength(ScanLen, fFFFieldCount+1);
                  while DefsFound < fFFFieldCount do
                  begin
                    inc(DefsFound);
                    ScanBest[DefsFound] := wffbs_Any;
                    ScanLen[DefsFound] := -1;
                    FldDefs.AddDef(Format('FIELD%4.4d',[DefsFound])).EditNoModify;
                  end;
                end;
              end else begin
              //!!!! Flag a warning?
              end;
            for fi := 1 to DefsFound do
            if fFFFieldLens[fi] > 0 then
            begin
              MaxFieldLen := ReadDelimitedField(@fFFRecBuffInfo, fi,fFFDelimQuotes[wffqOpenName], fFFDelimQuotes[wffqCloseName]);
              if MaxFieldLen > 0 then
              with FldDefs[fi-1] do
              begin
                if ScanLen[fi] < 0 then
                  ScanLen[fi] := MaxFieldLen
                else
                  if ScanLen[fi] <> MaxFieldLen then
                    ScanLen[fi] := 0;
                if MaxFieldLen > DSrcSize then
                  DSrcSize := MaxFieldLen;
                if ScanBest[fi] <> wffbs_Chars then
                  FldBuffBestType(MaxFieldLen, ScanBest[fi]);
              end else // if truedata > 0, with ...
                ScanLen[DefsFound] := 0;
            end else // for fi if basedata > 0
              ScanLen[DefsFound] := 0;
          end;  // for ri ....
          FldDefs.RecLen := 0;
          for fi := 1 to DefsFound do
          with FldDefs[fi-1] do
          begin
            case ScanBest[fi] of
              wffbs_Any   :
                begin
                  DSrcSize := 20;  // found no data - set a default
                end;
              wffbs_Chars :
                begin
                  if ScanLen[DefsFound] = 0 then // round length (not all same)
                    DSrcSize := 10 * Round((DSrcSize+5)/10);
                end;
              wffbs_Date  :
                begin
                  if ScanLen[DefsFound] = 0 then // round length (not all same)
                    DSrcSize := 10 * Round((DSrcSize+5)/10);
                  DUsgType := wffu_DateTime;
                  DSrcDstMask := ShortDateFormat + ' ' + LongTimeFormat;
                end;
              wffbs_Float :
                begin
                  DSrcType := wffs_NumUnPk;
                  DUsgType := wffu_Float;
                  DSrcPrec := DSrcSize;
                  DSrcSize := 20;
                end;
              wffbs_Integer :
                begin
                  DSrcType := wffs_NumUnPk;
                  DUsgType := wffu_LongInt;
                  DSrcSize := 10;
                end;
            end;
            FldDefs.RecLen := FldDefs.RecLen + DSrcSize;
          end;
          FldDefs.RecalcFiller;
        finally
          FreeData;
          DeAllocLayoutBuffers;
        end;
      finally
        for fi := DefsFound - 1 downto 0 do
          FldDefs[fi].Post;
      end;
    finally
      if not WasEditing then  FldDefs.Post;
    end;
  finally
    fDefinitions := SaveDefs;
  end;
end;

end.
