{
  Roland Potter - Woodstock Computer Solutions
  http://wcsweb.co.uk

  This code is provided for educatonal purposes only
  with no warranty whatsoever.

  Feel free to adapt it and use it in applcations commercial
  or otherwise and open source componenents but not as part
  of commercially distrbuted developer components.
}

// Use one of these in a multi level object structure that needs to be read
// and written to stream.
// Give the objects in the structure a TWStreamerMethod.
// Create a streamer and then call StreamObject.
// Within the method use only the Streamer's methods to read and write data.
//
// Versioning must be controlled externally but the idea should be
// to arrive at a stream that can be read and processed by newer OR older versions
// this is achieved by using read/write start and end procs that can skip to
// the correct flag type at the correct level of object nesting and thus ignore
// newly added items. A Version property is included to allow it to be easily
// passed down by the streamer to subobjects.
//
// Use the alternate verions with read/write/seek if a TStream is not available
// to work with.
unit wcsStreamer;

interface
uses
  Windows, SysUtils, Classes;
const
  WStreamBuffSize = $10000; // 64k
type
  TWStreamer = class;
  TWStreamOutMethod = function(const Buffer; Count: Longint): Longint of object;
  TWStreamInMethod  = function(var Buffer; Count: Longint): Longint of object;
  TWStreamSeekMethod= function(const Offset: Int64; Origin: TSeekOrigin): Int64 of object;
  TWStreamerMethod  = procedure(const Streamer: TWStreamer) of object;
  // This method must Add(Collection)/Create(List) the object when (not Streamer.Writing)
  // It will then normally call back to Streamer.StreamObject passing a TWStreamer Method for the
  // SubObject.
  TWStreamerSubMethod  = procedure(const Streamer: TWStreamer; var AnObject: TObject) of object;
  TObjectArray = array of TObject;

  TWStreamMark =
   (WStrmNull
   ,WStrmBreak
   ,WStrmObjectEnd
   ,WStrmEnd   // followed by BreakList, BreakListCount
   ,WStrmEnum
   ,WStrmChar
   ,WStrmByte
   ,WStrmWord1
   ,WStrmWord2
   ,WStrmLongint1
   ,WStrmLongint2
   ,WStrmLongint3
   ,WStrmLongint4
   ,WStrmInt641
   ,WStrmInt642
   ,WStrmInt643
   ,WStrmInt644
   ,WStrmInt645
   ,WStrmInt646
   ,WStrmInt647
   ,WStrmInt648
   ,WStrmDouble
   ,WStrmShortStr
   ,WStrmFixBuff1
   ,WStrmFixBuff2
   ,WStrmFixBuff3
   ,WStrmFixBuff4
   ,WStrmString1
   ,WStrmString2
   ,WStrmString3
   ,WStrmString4
   ,WStrmVarBuff1
   ,WStrmVarBuff2
   ,WStrmVarBuff3
   ,WStrmVarBuff4
   ,WStrmObject      // Flags start of something that ends with ObjectEnd!
   ,WStrmStringList
   ,WStrmObjectList  // uses callback to read/write/create objects
   ,WStrmStream      // same as VarBuff but reads/writes a stream.
   ,WStrmCollection  // uses callback to read/write/create objects
   );
  TWStreamMarks = Set of TWStreamMark;

  TWStreamBreak =  packed record
    wsb_ID  : Longint;
    wsb_Posn: Int64;
  end;

  TWStreamer = class
  private
    fWrite: TWStreamOutMethod;
    fRead : TWStreamInMethod;
    fSeek : TWStreamSeekMethod;
    fBuffer: array[0..WStreamBuffSize-1] of byte;
    fBreakCount: Longint;
    fBreakList : array of TWStreamBreak;
    fBuffPosn: Longint;
    fBuffLeft: Longint;
    fStreamRecursion: Byte;
    fWriting: Boolean;
    fStreamVers: integer;
    procedure Write(const Marker: TWStreamMark; const Output; const OutputLen: Longint);
    procedure FlushIn;
    procedure FlushOut;
    procedure BuffRead(const IntoData; const ReqLen: integer);
    procedure BuffSkip(const ReqLen: integer);
    procedure ReadFixBuff(var InputVar; const MaxLen: integer);
    procedure ReadFixed(const Markers: TWStreamMarks; var InputVar;
      const FixedLen: Integer);
    function ReadMarker(const Markers: TWStreamMarks): TWStreamMark;
    // Checks for Null or other marker alternative, rewinds if not null
    function CheckMarker(const Markers: TWStreamMarks): TWStreamMark;
    function ReadNonNullMarker(const Markers: TWStreamMarks): TWStreamMark;

    procedure ReadShortString(var InputVar);
    procedure ReadString(var InputVar: String);
    procedure SkipString;
    procedure ReadVarBuff(var InputPtr: Pointer; var InputLen: Integer);
  public
    constructor CreateForOutput(const WriteProc: TWStreamOutMethod;
                                const SeekProc: TWStreamSeekMethod);
    constructor CreateForInput (const ReadProc : TWStreamInMethod;
                                const SeekProc: TWStreamSeekMethod);
    constructor CreateTStreamOutput(const AStream: TStream);
    constructor CreateTStreamInput (const AStream: TStream);
    procedure StreamEnum(var Value);
    // streams as an enum and checks against a max (normally High(xxxType))
    // can support other historic streaming flags
    function StreamEnumProtect(var Value; const MaxVal, DefVal: Byte;
      const StreamMark: TWStreamMark = WStrmEnum): Boolean;
    procedure StreamChar(var Value: Char);
    procedure StreamByte(var Value: Byte);
    procedure StreamWord(var Value: Word);
    procedure StreamLongint(var Value: Longint);
    procedure StreamDWord(var Value: DWord);
    procedure StreamInt64(var Value: Int64);
    procedure StreamDouble(var Value: Double);
    procedure StreamShortString(var Value);
    procedure StreamString(var Value: String);
    procedure StreamFixBuff(var Value; const MaxLen: Longint);
    procedure StreamVarBuff(var ValuePtr: Pointer; var Length: Longint);
    procedure StreamStringList(const Value: TStrings);
    procedure StreamObject(const StreamerMethod: TWStreamerMethod);
    procedure StreamObjectList(const Values: TList;
      const StreamerMethod: TWStreamerSubMethod; const Limit: integer = MaxInt);
    procedure StreamCollection(const Values: TCollection;
      const StreamerSubMethod: TWStreamerSubMethod; const Limit: integer = MaxInt);
    procedure StreamObjectArray(var Values: TObjectArray; var Count: integer;
      const StreamerMethod: TWStreamerSubMethod; const Limit: integer = MaxInt);
    procedure StreamStream(const Stream: TStream);
    procedure StreamBreak(const ID: Longint);
    procedure StreamObjectStart;
    procedure StreamObjectEnd;
    procedure StreamEnd;
    procedure StreamVersion(var AVersion: Longint);

    property Writing: Boolean read fWriting;
    property StreamVers: integer read fStreamVers write fStreamVers;
  end;

implementation
const
  WStrmWords    = [WStrmWord1..WStrmWord2];
  WStrmLongints = [WStrmLongint1..WStrmLongint4];
  WStrmInt64s   = [WStrmInt641..WStrmInt648];
  WStrmFixBuffs = [WStrmFixBuff1..WStrmFixBuff4];
  WStrmStrings  = [WStrmString1..WStrmString4];
  WStrmVarBuffs = [WStrmVarBuff1..WStrmVarBuff4];
  WStrmLengths  = [WStrmShortStr..WStrmVarBuff4];

  WStreamMarkLen : array[TWStreamMark] of integer =
    (0 //WStrmNull
    ,0 //WStrmBreak
    ,0 //WStrmObjectEnd
    ,0 //WStrmEnd
    ,1 //WStrmEnum
    ,1 //WStrmChar
    ,1 //WStrmByte
    ,1 //WStrmWord1
    ,2 //WStrmWord2
    ,1 //WStrmLongint1
    ,2 //WStrmLongint2
    ,3 //WStrmLongint3
    ,4 //WStrmLongint4
    ,1 //WStrmInt641
    ,2 //WStrmInt642
    ,3 //WStrmInt643
    ,4 //WStrmInt644
    ,5 //WStrmInt645
    ,6 //WStrmInt646
    ,7 //WStrmInt647
    ,8 //WStrmInt648
    ,8 //WStrmDouble
    ,1 //WStrmShortStr
    ,1//WStrmFixBuff1
    ,1//WStrmFixBuff2
    ,1//WStrmFixBuff3
    ,1//WStrmFixBuff4
    ,1 //WStrmString1
    ,2 //WStrmString2
    ,3 //WStrmString3
    ,4 //WStrmString4
    ,1 //WStrmVarBuff1
    ,2 //WStrmVarBuff2
    ,3 //WStrmVarBuff3
    ,4 //WStrmVarBuff4
    ,0 //WStrmObject
    ,0 //WStrmStringList
    ,0 //WStrmObjectList
    ,0 //WStrmStream
    ,0 //WStrmCollection
    );
{ TWStreamer }

constructor TWStreamer.CreateForOutput(const WriteProc: TWStreamOutMethod;
  const SeekProc: TWStreamSeekMethod);
begin
  fWriting := True;
  fSeek  := SeekProc;
  fRead  := Nil;
  fWrite := WriteProc;
  fBuffLeft := WStreamBuffSize;
end;

constructor TWStreamer.CreateForInput(const ReadProc: TWStreamInMethod;
  const SeekProc: TWStreamSeekMethod);
begin
  fWriting := False;
  fSeek  := SeekProc;
  fRead  := ReadProc;
  fWrite := Nil;
end;

constructor TWStreamer.CreateTStreamOutput(const AStream: TStream);
begin
  CreateForOutput(AStream.Write, AStream.Seek);
end;

constructor TWStreamer.CreateTStreamInput(const AStream: TStream);
begin
  CreateForInput(AStream.Read, AStream.Seek);
end;

procedure TWStreamer.FlushOut;
begin
  if fBuffPosn > 0 then fWrite(fBuffer[0], fBuffPosn);
  fBuffPosn := 0;
  fBuffLeft := WStreamBuffSize;
end;

procedure TWStreamer.FlushIn;
begin
  fBuffLeft := fRead(fBuffer[0], WStreamBuffSize);
  fBuffPosn := 0;
end;

procedure TWStreamer.Write(const Marker: TWStreamMark; const Output; const OutputLen: Integer);
var
  OutLeft : integer;
  OutPtr: PByte;
begin
  fBuffer[fBuffPosn] := Byte(Marker);
  inc(fBuffPosn);
  dec(fBuffLeft);
// Write the markers
  if Marker in WStrmLengths then
  begin
    OutLeft := WStreamMarkLen[Marker];
    if fBuffLeft < OutLeft then Flushout;
    move(OutPutLen, fBuffer[fBuffPosn], OutLeft);
    inc(fBuffPosn, OutLeft);
    dec(fBuffLeft, OutLeft);
  end;

  OutLeft := OutputLen;
  OutPtr  := @Output;
// >= here ensures it never enters here with a full buffer.
  while OutLeft >= fBuffLeft do
  begin
    if fBuffLeft > 0 then
    begin
      move(OutPtr^, fBuffer[fBuffPosn], fBuffLeft);
      inc(fBuffPosn, fBuffLeft);
      dec(OutLeft, fBuffLeft);
      inc(OutPtr, fBuffLeft);
    end;
    FlushOut;
  end;
  if OutLeft > 0 then
  begin
    move(OutPtr^, fBuffer[fBuffPosn], OutLeft);
    inc(fBuffPosn, OutLeft);
    dec(fBuffLeft, OutLeft);
  end;
end;

procedure TWStreamer.BuffRead(const IntoData; const ReqLen: integer);
var
  ReqLeft : integer;
  ReqData : PChar;
begin
  ReqData := PChar(@IntoData);
  ReqLeft := ReqLen;
  while ReqLeft > 0 do
  begin
    if fBuffLeft = 0 then FlushIn;
    if fBuffLeft = 0 then raise exception.Create('Read beyond end of Stream');
    if fBuffLeft < ReqLeft then
    begin
      move(fBuffer[fBuffPosn], ReqData[0], fBuffLeft);
      dec(ReqLeft, fBuffLeft);
      inc(ReqData, fBuffLeft);
      fBuffLeft := 0;
    end else begin
      move(fBuffer[fBuffPosn], ReqData[0], ReqLeft);
      inc(fBuffPosn, ReqLeft);
      dec(fBuffLeft, ReqLeft);
      ReqLeft := 0;
    end;
  end;
end;

procedure TWStreamer.BuffSkip(const ReqLen: integer);
var
  ReqLeft : integer;
begin
  ReqLeft := ReqLen;
  while ReqLeft > 0 do
  begin
    if fBuffLeft = 0 then FlushIn;
    if fBuffLeft = 0 then raise exception.Create('Read beyond end of Stream');
    if fBuffLeft < ReqLeft then
    begin
      dec(ReqLeft, fBuffLeft);
      fBuffLeft := 0;
    end else begin
      inc(fBuffPosn, ReqLeft);
      dec(fBuffLeft, ReqLeft);
      ReqLeft := 0;
    end;
  end;
end;

function TWStreamer.ReadMarker(const Markers: TWStreamMarks): TWStreamMark;
begin
  BuffRead(Result, 1);
  while Result = WStrmBreak do BuffRead(Result, 1);
  if Result = WStrmNull then Exit;
  if not (Result in Markers) then
    raise exception.Create('Invalid Stream Marker');
end;

function TWStreamer.CheckMarker(const Markers: TWStreamMarks): TWStreamMark;
begin
  BuffRead(Result, 1);
  while Result = WStrmBreak do BuffRead(Result, 1);
  if Result = WStrmNull then Exit;
  if not (Result in Markers) then
    raise exception.Create('Invalid Stream Marker');
  dec(fBuffPosn);
  inc(fBuffLeft);
end;

function TWStreamer.ReadNonNullMarker(const Markers: TWStreamMarks): TWStreamMark;
begin
  BuffRead(Result, 1);
  while Result = WStrmBreak do BuffRead(Result, 1);
  if not (Result in Markers) then
    raise exception.Create('Invalid Stream Marker');
end;

procedure TWStreamer.ReadString(var InputVar: String);
var
  BuffToRead: integer;
  ReadMark: TWStreamMark;
begin
  ReadMark := ReadMarker(WStrmStrings);
  if ReadMark = WStrmNull then
    InputVar := ''
  else begin
    BuffToRead := 0;
    BuffRead(BuffToRead, WStreamMarkLen[ReadMark]);
    if BuffToRead > 0 then
    begin
      SetLength(InputVar, BuffToRead);
      BuffRead(InputVar[1], BuffToRead);
    end else
      InputVar := ''
  end;
end;

procedure TWStreamer.SkipString;
var
  BuffToRead: integer;
  ReadMark : TWStreamMark;
begin
  ReadMark := ReadMarker(WStrmStrings);
  if ReadMark  <> WStrmNull then
  begin
    BuffToRead := 0;
    BuffRead(BuffToRead, WStreamMarkLen[ReadMark]);
    if BuffToRead > 0 then BuffSkip(BuffToRead);
  end;
end;

procedure TWStreamer.ReadShortString(var InputVar);
var
  BuffToRead: Byte;
begin
  if ReadMarker([WStrmShortStr]) = WStrmNull then
    Byte(InputVar) := 0
  else begin
    BuffRead(BuffToRead, 1);
    Byte(InputVar) := BuffToRead;
    if BuffToRead > 0 then
      BuffRead(PChar(@InputVar)[1], BuffToRead);
  end;
end;

procedure TWStreamer.ReadFixBuff(var InputVar; const MaxLen: integer);
var
  BuffToRead: integer;
  ReadMark: TWStreamMark;
begin
  ReadMark := ReadMarker(WStrmFixBuffs);
  BuffToRead := 0;
  if ReadMark <> WStrmNull then
    BuffRead(BuffToRead, WStreamMarkLen[ReadMark]);
  if BuffToRead > MaxLen then
  begin
    BuffRead(InputVar, MaxLen);
    BuffSkip(BuffToRead-MaxLen);
  end else
    if BuffToRead < MaxLen then
    begin
      FillChar(PChar(@InputVar)[BuffToRead], MaxLen-BuffToRead, 0);
      if BuffToRead > 0 then BuffRead(InputVar, BuffToRead);
    end else
      BuffRead(InputVar, BuffToRead);
end;

procedure TWStreamer.ReadVarBuff(var InputPtr: Pointer; var InputLen: Integer);
var
  ReadMark: TWStreamMark;
begin
  InputPtr := Nil;
  InputLen := 0;
  if ReadMarker(WStrmVarBuffs) <> WStrmNull then
  begin
    InputLen := 0;
    BuffRead(InputLen, WStreamMarkLen[ReadMark]);
    if InputLen > 0 then
    begin
      GetMem(InputPtr, InputLen);
      BuffRead(InputPtr^, InputLen);
    end;
  end;
end;

procedure TWStreamer.ReadFixed(const Markers: TWStreamMarks; var InputVar;
  const FixedLen: Integer);
var
  BuffReq: integer;
  ReadMark: TWStreamMark;
begin
  ReadMark:= ReadMarker(Markers);
  if ReadMark = WStrmNull then
    FillChar(InputVar, FixedLen, 0)
  else begin
    BuffReq := WStreamMarkLen[ReadMark];
    if BuffReq < FixedLen then // zero the hi part
      FillChar(PChar(@InputVar)[BuffReq], FixedLen - BuffReq, 0);
    BuffRead(InputVar, BuffReq);
  end;
end;

procedure TWStreamer.StreamFixBuff(var Value; const MaxLen: Longint);
begin
  if fWriting then
  begin
    if (MaxLen and $FFFFFF00) = 0 then Write(WStrmFixBuff1, Value, MaxLen)
    else if (MaxLen and $FFFF0000) = 0 then Write(WStrmFixBuff2, Value, MaxLen)
    else if (MaxLen and $FF000000) = 0 then Write(WStrmFixBuff3, Value, MaxLen)
    else Write(WStrmFixBuff4, Value, MaxLen);
  end else
    ReadFixBuff(Value, MaxLen);
end;

procedure TWStreamer.StreamVarBuff(var ValuePtr: Pointer; var Length: Longint);
begin
  if fWriting then
  begin
    if Length = 0 then Write(WStrmNull, ValuePtr^, 0)
    else if (Length and $FFFFFF00) = 0 then Write(WStrmVarBuff1, ValuePtr^, Length)
    else if (Length and $FFFF0000) = 0 then Write(WStrmVarBuff2, ValuePtr^, Length)
    else if (Length and $FF000000) = 0 then Write(WStrmVarBuff3, ValuePtr^, Length)
    else Write(WStrmVarBuff4, ValuePtr^, Length);
  end else
    ReadVarBuff(ValuePtr, Length);
end;

procedure TWStreamer.StreamString(var Value: String);
var
  L : integer;
begin
  if fWriting then
  begin
    L := Length(Value);
    if L = 0 then Write(WStrmNull, Value, 0)
    else if (L and $FFFFFF00) = 0 then Write(WStrmString1, Value[1], L)
    else if (L and $FFFF0000) = 0 then Write(WStrmString2, Value[1], L)
    else if (L and $FF000000) = 0 then Write(WStrmString3, Value[1], L)
    else Write(WStrmString4, Value[1], L);
  end else
    ReadString(Value);
end;

procedure TWStreamer.StreamShortString(var Value);
begin
  if fWriting then
  begin
    if Byte(Value) = 0 then Write(WStrmNull, Value, 0)
    else Write(WStrmShortStr, Value, Byte(Value)+1);
  end else
    ReadShortString(Value);
end;

procedure TWStreamer.StreamByte(var Value: Byte);
begin
  if fWriting then
    if Value > 0 then
      Write(WStrmByte, Value, 1)
    else
      Write(WStrmNull, Value, 0)
  else
    ReadFixed([WStrmByte], Value, 1);
end;

procedure TWStreamer.StreamChar(var Value: Char);
begin
  if fWriting then
    if Value > #0 then Write(WStrmChar, Value, 1)
    else Write(WStrmNull, Value, 0)
  else
    ReadFixed([WStrmChar], Value, 1);
end;

procedure TWStreamer.StreamEnum(var Value);
begin
  if fWriting then
    if Byte(Value) > 0 then Write(WStrmENum, Value, 1)
    else Write(WStrmNull, Value, 0)
  else
    ReadFixed([WStrmENum], Value, 1);
end;

function TWStreamer.StreamEnumProtect(var Value; const MaxVal, DefVal: Byte;
  const StreamMark: TWStreamMark = WStrmEnum): Boolean;
var
  StreamVal: Byte;
begin
  if fWriting then
  begin
    StreamVal:= Byte(Value);
    Result := StreamVal <= MaxVal;
    if not Result then StreamVal := DefVal;
    if StreamVal > 0 then
      Write(StreamMark, StreamVal, 1)
    else
      Write(WStrmNull, Value, 0);
  end else begin
    ReadFixed([StreamMark], StreamVal, 1);
    Result := StreamVal <= MaxVal;
    if Result then
      Byte(Value) := StreamVal
    else
      Byte(Value) := DefVal;
  end;
end;

procedure TWStreamer.StreamWord(var Value: Word);
begin
  if fWriting then
  begin
    if Value = 0 then Write(WStrmNull, Value, 0)
    else if (Value and $FF00) = 0 then Write(WStrmWord1, Value, 1)
    else Write(WStrmWord2, Value, 2);
  end else
    ReadFixed(WStrmWords, Value, 2);
end;

procedure TWStreamer.StreamLongint(var Value: Integer);
begin
  if fWriting then
  begin
    if Value = 0 then Write(WStrmNull, Value, 0)
    else if (Value and $FFFFFF00) = 0 then Write(WStrmLongint1, Value, 1)
    else if (Value and $FFFF0000) = 0 then Write(WStrmLongint2, Value, 2)
    else if (Value and $FF000000) = 0 then Write(WStrmLongint3, Value, 3)
    else Write(WStrmLongint4, Value, 4);
  end else
    ReadFixed(WStrmLongints, Value, 4);
end;

procedure TWStreamer.StreamDWord(var Value: DWord);
begin
  StreamLongint(Longint(Value));
end;

procedure TWStreamer.StreamInt64(var Value: Int64);
begin
  if fWriting then
  begin
    if Value = 0 then Write(WStrmNull, Value, 0)
    else if (Value and $FFFFFFFFFFFFFF00) = 0 then Write(WStrmInt641, Value, 1)
    else if (Value and $FFFFFFFFFFFF0000) = 0 then Write(WStrmInt642, Value, 2)
    else if (Value and $FFFFFFFFFF000000) = 0 then Write(WStrmInt643, Value, 3)
    else if (Value and $FFFFFFFF00000000) = 0 then Write(WStrmInt644, Value, 4)
    else if (Value and $FFFFFF0000000000) = 0 then Write(WStrmInt645, Value, 5)
    else if (Value and $FFFF000000000000) = 0 then Write(WStrmInt646, Value, 6)
    else if (Value and $FF00000000000000) = 0 then Write(WStrmInt647, Value, 7)
    else Write(WStrmInt648, Value, 8);
  end else
    ReadFixed(WStrmInt64s, Value, 8);
end;

procedure TWStreamer.StreamDouble(var Value: Double);
begin
  if fWriting then
    if Value <> 0 then Write(WStrmDouble, Value, 8)
    else Write(WStrmNull, Value, 0)
  else
    ReadFixed([WStrmDouble], Value, 8);
end;

procedure TWStreamer.StreamObjectEnd;
var
  ReadMark: TWStreamMark;
  BuffReq, i: integer;
begin
  if fWriting then
    Write(WStrmObjectEnd, Self, 0)
  else begin
    inc(fStreamRecursion);
    if fStreamRecursion > 64 then
      raise exception.Create('Excessive Streamer EndMark Recursion');
    try
      BuffRead(ReadMark, 1);
      while ReadMark <> WStrmObjectEnd do
      begin
        if ReadMark in WStrmLengths then
        begin
          BuffReq:= 0;
          BuffRead(BuffReq, WStreamMarkLen[ReadMark]);
          if BuffReq > 0 then BuffSkip(BuffReq);
        end else if ReadMark = WStrmEnd then
          raise exception.Create('End of Stream Marker encountered')
        else if WStreamMarkLen[ReadMark] > 0 then
          BuffSkip(WStreamMarkLen[ReadMark])
        else case ReadMark of
          WStrmObject     : StreamObjectEnd; // Recursive!
          WStrmStringList : begin
                              StreamLongint(BuffReq);
                              for i := 0 to BuffReq-1 do SkipString;
                            end;
          WStrmObjectList : StreamLongint(BuffReq); // now just pick up object starts and ends
          WStrmCollection : StreamLongint(BuffReq); // now just pick up object starts and ends
          WStrmStream     : begin
                              StreamLongint(BuffReq);
                              BuffSkip(BuffReq);
                            end;
        end;
        BuffRead(ReadMark, 1);
      end;
    finally
      dec(fStreamRecursion);
    end;
  end;
end;

procedure TWStreamer.StreamEnd;
var
  Marker: TWStreamMark;
begin
  if fWriting then
  begin
    FlushOut;
    Marker := WStrmEnd;
    fWrite(Marker,1);
    if fBreakCount > 0 then
      fWrite(fBreakList[0], fBreakCount*Sizeof(TWStreamBreak));
    fWrite(fBreakCount, 4);
  end else begin
    BuffRead(Marker, 1);
    if Marker <> WStrmEnd then
      raise exception.Create('Invalid Stream End Marker');
  end;
end;

procedure TWStreamer.StreamBreak(const ID: Integer);
var
  i : integer;
  Marker: TWStreamMark;
begin
  if fWriting then
  begin
    // flush the buffer - add to the Break table
    if fBuffPosn > 0 then FlushOut;
    inc(fBreakCount);
    SetLength(fBreakList, fBreakCount);
    fBreakList[fBreakCount-1].wsb_ID := ID;
    fBreakList[fBreakCount-1].wsb_Posn := fSeek(0, soCurrent);
    Write(WStrmBreak, Self, 0);
  end else begin
    if fBreakCount = 0 then
    begin
      fSeek(-4, soEnd);
      fRead(fBreakCount, 4);
      SetLength(fBreakList, fBreakCount);
      fSeek(-(5 + fBreakCount*Sizeof(TWStreamBreak)), soEnd);
      FlushIn;
      BuffRead(Marker, 1);
      if Marker <> wStrmEnd then
        raise exception.Create('Invalid Stream End Marker');
      if fBreakCount > 0 then
        BuffRead(fBreakList[0], fBreakCount*Sizeof(TWStreamBreak));
    end;
    for i := 0 to fBreakCount - 1 do
      if fBreakList[i].wsb_ID = ID then
      begin
        if fSeek(fBreakList[i].wsb_Posn, soBeginning) = fBreakList[i].wsb_Posn then
        begin
          FlushIn;
          BuffRead(Marker, 1);
          if Marker <> wStrmBreak then
            raise exception.Create('Invalid Stream Break Marker');
        end else
          raise Exception.Createfmt('Error seeking to Break ID [%d] Position [%d]',
                                    [ID, fBreakList[i].wsb_Posn]);
        Exit;
      end;
    raise Exception.Createfmt('Break ID [%d] not found', [ID]);
  end;
end;

procedure TWStreamer.StreamStringList(const Value: TStrings);
var
  i,c : integer;
  s : string;
begin
  if fWriting then
  begin
    c := Value.Count;
    if c > 0 then
    begin
      Write(WStrmStringList, c, 0);
      StreamLongint(c);
      for i := 0 to c - 1 do
      begin
        S := Value[i];
        StreamString(S);
      end;
    end else
      Write(WStrmNull, c, 0);
  end else begin
    Value.Clear;
    if ReadMarker([WStrmStringList]) <> WStrmNull then
    begin
      StreamLongint(c);
      for i := 0 to c - 1 do
      begin
        StreamString(S);
        Value.Add(S);
      end;
    end;
  end;
end;

procedure TWStreamer.StreamObjectStart;
begin
  Write(WStrmObject, Self, 0);
end;

procedure TWStreamer.StreamObject(const StreamerMethod: TWStreamerMethod);
begin
  if fWriting then
    Write(WStrmObject, Self, 0)
  else
    ReadNonNullMarker([WStrmObject]);
  StreamerMethod(Self);
  StreamObjectEnd;
end;

procedure TWStreamer.StreamObjectList(const Values: TList;
  const StreamerMethod: TWStreamerSubMethod; const Limit: integer = MaxInt);
var
  i,c : integer;
  ReadMark: TWStreamMark;
  Obj: TObject;
begin
  if fWriting then
  begin
    c := Values.Count;
    if c > 0 then
    begin
      Write(WStrmObjectList, c, 0);
      StreamLongint(c);
      for i := 0 to c - 1 do
      begin
        Obj := Values[i];
        StreamerMethod(Self, Obj);
      end;
    end else
      Write(WStrmNull, c, 0);
  end else begin
    Values.Clear;
    if ReadMarker([WStrmObjectList]) <> WStrmNull then
    begin
      StreamLongint(c);
      if c > Limit then
        raise exception.Create('Excessive Streamer SubObject Count: '+IntToStr(c)+' vs '+IntToStr(Limit));
      for i := 0 to c - 1 do
      begin
        StreamerMethod(Self, Obj);
        Values.Add(Obj);
      end;
    end;
  end;
end;

procedure TWStreamer.StreamObjectArray(var Values: TObjectArray; var Count: integer;
  const StreamerMethod: TWStreamerSubMethod; const Limit: integer = MaxInt);
var
  i : integer;
  ReadMark: TWStreamMark;
  Obj: TObject;
begin
  if fWriting then
  begin
    if count > 0 then
    begin
      Write(WStrmObjectList, count, 0);
      StreamLongint(count);
      for i := 0 to count - 1 do
      begin
        Obj := Values[i];
        if Obj <> Nil then
          StreamerMethod(Self, Obj)
        else
          Write(WStrmNull, count, 0);
      end;
    end else
      Write(WStrmNull, count, 0);
  end else begin
    for i := 0 to Length(Values) - 1 do
      if Values[i] <> Nil then Values[i].Free;
    SetLength(Values, 0);
    if ReadMarker([WStrmObjectList]) <> WStrmNull then
    begin
      StreamLongint(count);
      if count > Limit then
        raise exception.Create('Excessive Streamer SubObject Count');
      SetLength(Values, count);
      FillChar(Values[0], count * Sizeof(TObject), 0);
      for i := 0 to count - 1 do
        if CheckMarker([WStrmObject]) <> WStrmNull then
        begin
          StreamerMethod(Self, Obj);
          Values[i] := Obj;
        end;
    end;
  end;
end;

procedure TWStreamer.StreamStream(const Stream: TStream);
var
  i,c : integer;
begin
  if fWriting then
  begin
    c := Stream.Size;
    if c > 0 then
    begin
      Write(WStrmStream, c, 0);
      StreamLongint(c);
      Stream.Position := 0;

      while c > 0 do
      begin
        if c >= fBuffLeft then
        begin
          Stream.Read(fBuffer[fBuffPosn], fBuffLeft);
          inc(fBuffPosn, fBuffLeft);
          dec(c, fBuffLeft);
          fBuffLeft := 0;
          Flushout;
        end else begin
          Stream.Read(fBuffer[fBuffPosn], c);
          inc(fBuffPosn, c);
          dec(fBuffLeft, c);
          c := 0;
        end;
      end;
    end else
      Write(WStrmNull, c, 0);
  end else begin
    if ReadMarker([WStrmStream]) <> WStrmNull then
    begin
      StreamLongint(c);
      Stream.Size := c;
      Stream.Position := 0;
      while c > 0 do
      begin
        if fBuffLeft = 0 then FlushIn;
        if fBuffLeft = 0 then raise exception.Create('Read beyond end of Stream');
        if c >= fBuffLeft then
        begin
          Stream.Write(fBuffer[fBuffPosn], fBuffLeft);
          inc(fBuffPosn, fBuffLeft);
          dec(c, fBuffLeft);
          fBuffLeft := 0;
        end else begin
          Stream.Write(fBuffer[fBuffPosn], c);
          inc(fBuffPosn, c);
          dec(fBuffLeft, c);
          c := 0;
        end;
      end;
      if fBuffLeft = 0 then FlushIn;
      Stream.Position := 0;
    end else
      Stream.Size := 0;
  end;
end;

procedure TWStreamer.StreamCollection(const Values: TCollection;
  const StreamerSubMethod: TWStreamerSubMethod; const Limit: integer);
var
  i,c : integer;
  ReadMark: TWStreamMark;
  Obj: TObject;
begin
  if fWriting then
  begin
    c := Values.Count;
    if c > 0 then
    begin
      Write(WStrmCollection, c, 0);
      StreamLongint(c);
      for i := 0 to c - 1 do
      begin
        Obj := Values.Items[i];
        StreamerSubMethod(Self, Obj);
      end;
    end else
      Write(WStrmNull, c, 0);
  end else begin
    Values.Clear;
    if ReadMarker([wStrmCollection]) <> WStrmNull then
    begin
      StreamLongint(c);
      if c > Limit then
        raise exception.Create('Excessive Streamer SubObject Count');
      for i := 0 to c - 1 do
        StreamerSubMethod(Self, Obj);
    end;
  end;
end;

procedure TWStreamer.StreamVersion(var AVersion: Longint);
begin
  if Writing then fStreamVers := AVersion;
  StreamLongint(fStreamVers);
  if not Writing then AVersion := fStreamVers;
end;

end.
